from __future__ import annotations

from typing import Sequence

from dbt_semantic_interfaces.implementations.saved_query import PydanticSavedQuery, PydanticSavedQueryQueryParams
from dbt_semantic_interfaces.references import EntityReference

from metricflow_semantics.naming.object_builder_scheme import ObjectBuilderNamingScheme
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.test_helpers.synthetic_manifest.categorical_dimension_generator import (
    CategoricalDimensionGenerator,
)
from metricflow_semantics.test_helpers.synthetic_manifest.metric_generator import MetricGenerator
from metricflow_semantics.test_helpers.synthetic_manifest.synthetic_manifest_parameter_set import (
    SyntheticManifestParameterSet,
)


class SavedQueryGenerator:
    """Helps generate saved queries for the synthetic manifest."""

    def __init__(  # noqa: D107
        self,
        parameter_set: SyntheticManifestParameterSet,
        metric_generator: MetricGenerator,
        categorical_dimension_generator: CategoricalDimensionGenerator,
    ) -> None:
        self._parameter_set = parameter_set
        self._metric_generator = metric_generator
        self._dimension_generator = categorical_dimension_generator
        self._naming_scheme = ObjectBuilderNamingScheme()

    def _get_saved_query_name(self, saved_query_index: int) -> str:
        return f"saved_query_{saved_query_index:03}"

    def generate_saved_queries(self) -> Sequence[PydanticSavedQuery]:  # noqa: D102
        saved_queries = []
        next_metric_index = self._metric_generator.get_first_index_at_max_depth()
        next_categorical_dimension_index = 0

        for saved_query_index in range(self._parameter_set.saved_query_count):
            metrics = []
            for _ in range(self._parameter_set.metrics_per_saved_query):
                metrics.append(self._metric_generator.get_metric_name(next_metric_index))
                next_metric_index = self._metric_generator.get_next_wrapped_width_index(next_metric_index)
            categorical_dimensions = []
            for _ in range(self._parameter_set.categorical_dimensions_per_saved_query):
                categorical_dimensions.append(
                    self._naming_scheme.input_str(
                        DimensionSpec(
                            element_name=self._dimension_generator.get_dimension_name(next_categorical_dimension_index),
                            entity_links=(EntityReference(self._parameter_set.common_entity_name),),
                        )
                    )
                )
                next_categorical_dimension_index = self._dimension_generator.get_next_wrapped_index(
                    next_categorical_dimension_index
                )

            saved_queries.append(
                PydanticSavedQuery(
                    name=self._get_saved_query_name(saved_query_index),
                    query_params=PydanticSavedQueryQueryParams(
                        metrics=metrics,
                        group_by=categorical_dimensions,
                    ),
                )
            )

        return saved_queries



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from dbt_semantic_interfaces.references import TimeDimensionReference
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class MetricTimeDimensionTransformNode(DataflowPlanNode):
    """A node transforms the input data set so that it contains the metric time dimension and relevant simple-metric inputs.

    The metric time dimension is used later to aggregate all simple-metric inputs in the data set.

    Input: a data set containing simple-metric inputs along with the associated aggregation time dimension.

    Output: a data set similar to the input data set, but includes the configured aggregation time dimension as the
    metric time dimension and only contains simple-metric inputs that are defined to use it.

    Attributes:
        aggregation_time_dimension_reference: The time dimension that simple-metric inputs in the input should be aggregated to.
    """

    aggregation_time_dimension_reference: TimeDimensionReference

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        aggregation_time_dimension_reference: TimeDimensionReference,
    ) -> MetricTimeDimensionTransformNode:
        return MetricTimeDimensionTransformNode(
            parent_nodes=(parent_node,),
            aggregation_time_dimension_reference=aggregation_time_dimension_reference,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_SET_METRIC_AGGREGATION_TIME

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_metric_time_dimension_transform_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"Metric Time Dimension '{self.aggregation_time_dimension_reference.element_name}'"

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (
            DisplayedProperty("aggregation_time_dimension", self.aggregation_time_dimension_reference.element_name),
        )

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.aggregation_time_dimension_reference == self.aggregation_time_dimension_reference
        )

    def with_new_parents(  # noqa: D102
        self, new_parent_nodes: Sequence[DataflowPlanNode]
    ) -> MetricTimeDimensionTransformNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return MetricTimeDimensionTransformNode.create(
            parent_node=new_parent_nodes[0],
            aggregation_time_dimension_reference=self.aggregation_time_dimension_reference,
        )



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Sequence, Set
from dataclasses import dataclass
from typing import Generic, Optional, Sized, TypeVar

from typing_extensions import Self, override

from metricflow_semantics.experimental.mf_graph.comparable import Comparable, ComparisonKey
from metricflow_semantics.experimental.mf_graph.mf_graph import MetricFlowGraphEdge, MetricFlowGraphNode
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext
from metricflow_semantics.mf_logging.pretty_print import mf_pformat

logger = logging.getLogger(__name__)


NodeT = TypeVar("NodeT", bound=MetricFlowGraphNode)
EdgeT = TypeVar("EdgeT", bound=MetricFlowGraphEdge)


class MetricFlowGraphPath(Generic[NodeT, EdgeT], Comparable, MetricFlowPrettyFormattable, Sized, ABC):
    """A read-only interface that describes a path in a directed graph."""

    @property
    @abstractmethod
    def edges(self) -> Sequence[EdgeT]:
        """The edges (in order) that constitute this path."""
        raise NotImplementedError

    @property
    @abstractmethod
    def nodes(self) -> Sequence[NodeT]:
        """The nodes (in order) that constitute this path."""
        raise NotImplementedError

    @property
    @abstractmethod
    def weight(self) -> int:
        """The weight of this path as defined by the weight function that was used during traversal."""
        raise NotImplementedError

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        # return format_context.formatter.pretty_format([node.node_descriptor.node_name for node in self.nodes])
        return format_context.formatter.pretty_format_object_by_parts(
            class_name=self.__class__.__name__,
            field_mapping={
                "nodes": [node.node_descriptor.node_name for node in self.nodes],
                "weight": self.weight,
            },
        )

    @property
    @abstractmethod
    def node_set(self) -> Set[NodeT]:
        """The set of nodes in this path. Useful for fast checks for cycles during path extension."""
        raise NotImplementedError()

    @override
    def __len__(self) -> int:
        return len(self.nodes)

    @abstractmethod
    def copy(self) -> Self:
        """Return a shallow copy of this path."""
        raise NotImplementedError

    def arrow_format(self) -> str:
        """Return a string representation that uses `->` between nodes.

        TODO: This is only used in tests so it should be migrated elsewhere.
        """
        return f"[weight: {self.weight}] " + " -> ".join([mf_pformat(node) for node in self.nodes])


@dataclass
class MutableGraphPath(MetricFlowGraphPath[NodeT, EdgeT], Generic[NodeT, EdgeT]):
    """A path that can be extended and also reverted back to the previous state before extension.

    * This is a mutable class as path-finding can traverse many edges, and using a single mutable object reduces
      overhead.
    * The append / pop functionality is useful for DFS traversal.
    * When an edge is added to the path, the incremental weight added by the edge is specified by the caller (this
      does not do any weight calculation).
    """

    _nodes: list[NodeT]
    _edges: list[EdgeT]
    _current_weight: int
    _current_node_set: set[NodeT]

    # As this path is extended step-by-step, keep track of the weights added so that when `pop()` is called, the weight
    # of the path afterward can be easily computed by subtracting the last incremental weight added. Similar situation
    # for `_node_set_addition_order`.
    _weight_addition_order: list[int]
    _node_set_addition_order: list[Optional[NodeT]]

    @staticmethod
    def create(start_node: Optional[NodeT] = None) -> MutableGraphPath:  # noqa: D102
        path: MutableGraphPath[NodeT, EdgeT] = MutableGraphPath(
            _nodes=[],
            _edges=[],
            _weight_addition_order=[],
            _current_weight=0,
            _current_node_set=set(),
            _node_set_addition_order=[],
        )
        if start_node:
            path._append_node(start_node)
        return path

    @property
    def edges(self) -> Sequence[EdgeT]:  # noqa: D102
        return self._edges

    @property
    def nodes(self) -> Sequence[NodeT]:  # noqa: D102
        return self._nodes

    @property
    def is_empty(self) -> bool:  # noqa: D102
        return not self._nodes

    def _append_node(self, node: NodeT) -> None:
        """Helper to add a node to the path."""
        self._nodes.append(node)
        if node in self._current_node_set:
            self._node_set_addition_order.append(None)
        else:
            self._current_node_set.add(node)
            self._node_set_addition_order.append(node)

    def append_edge(self, edge: EdgeT, weight: int) -> None:
        """Add an edge with the given weight to this path."""
        tail_node = edge.tail_node
        head_node = edge.head_node
        if len(self._nodes) == 0:
            self._append_node(tail_node)
        self._append_node(head_node)
        self._edges.append(edge)
        self._weight_addition_order.append(weight)
        self._current_weight += weight

    @property
    def weights(self) -> Sequence[int]:
        """The sequence of weights that were added to compute the total weight of the path."""
        return self._weight_addition_order

    @property
    def weight(self) -> int:
        """The current weight of the path."""
        return self._current_weight

    def pop_end(self) -> None:
        """Remove the last node / edge added to the path."""
        if not self._edges:
            if not self._nodes:
                raise KeyError("Can't pop an empty path")
            self._nodes.pop()
            return

        self._edges.pop()
        self._nodes.pop()
        weight = self._weight_addition_order.pop()
        self._current_weight -= weight

        added_node = self._node_set_addition_order.pop()
        if added_node is not None:
            self._current_node_set.remove(added_node)
        return

    @property
    def node_set(self) -> Set[NodeT]:
        """A set containing the nodes in this path."""
        return self._current_node_set

    @override
    def copy(self) -> Self:
        # noinspection PyArgumentList
        return self.__class__(
            _nodes=self._nodes.copy(),
            _edges=self._edges.copy(),
            _current_weight=self._current_weight,
            _current_node_set=self._current_node_set.copy(),
            _weight_addition_order=self._weight_addition_order.copy(),
            _node_set_addition_order=self._node_set_addition_order.copy(),
        )

    @property
    @override
    def comparison_key(self) -> ComparisonKey:
        return (
            tuple(self._nodes),
            self._current_weight,
            tuple(self._edges),
        )


MutablePathT = TypeVar("MutablePathT", bound=MutableGraphPath)



from __future__ import annotations

import logging
from typing import Optional

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class MetricFlowGraphNodeDescriptor:
    """Descriptor for a node to allow for lookups by strings."""

    node_name: str
    # A name to cluster nodes in a graph. Currently only used for visualization.
    cluster_name: Optional[str]



from __future__ import annotations

from typing import Optional

from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow_semantics.errors.error_classes import FeatureNotSupportedError


def error_if_not_standard_grain(input_granularity: str, context: Optional[str] = None) -> TimeGranularity:
    """Cast input grainularity string to TimeGranularity, otherwise error.

    TODO: Not needed once, custom grain is supported for most things.
    """
    try:
        time_grain = TimeGranularity(input_granularity)
    except ValueError:
        error_msg = f"Received a non-standard time granularity, which is not supported at the moment, received: {input_granularity}."
        if context:
            error_msg += f"\nContext: {context}"
        raise FeatureNotSupportedError(error_msg)
    return time_grain



from __future__ import annotations

import logging
from enum import Enum

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols import Metric
from dbt_semantic_interfaces.type_enums import MetricType
from typing_extensions import override

from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.metricflow_exception import InvalidManifestException
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.builder.subgraph_generator import (
    SemanticSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.edges.edge_labels import (
    CumulativeMetricLabel,
    DenyDatePartLabel,
    DenyEntityKeyQueryResolutionLabel,
    DenyVisibleAttributesLabel,
)
from metricflow_semantics.experimental.semantic_graph.edges.sg_edges import MetricDefinitionEdge
from metricflow_semantics.experimental.semantic_graph.nodes.entity_nodes import (
    ComplexMetricNode,
    MetricNode,
    SimpleMetricNode,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraphEdge,
    SemanticGraphNode,
)
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.metric_lookup import MetricLookup

logger = logging.getLogger(__name__)


class _SpecialCase(Enum):
    """Enumerates the special cases that affect the available group-by items for a metric."""

    CONVERSION_INPUT_METRIC = "conversion_input_metric"
    CUMULATIVE_METRIC = "cumulative_metric"
    CUMULATIVE_METRIC_WITH_WINDOW_OR_GRAIN_TO_DATE = "cumulative_metric_with_window_or_grain_to_date"
    TIME_OFFSET_DERIVED_METRIC = "time_offset_derived_metric"


class ComplexMetricSubgraphGenerator(SemanticSubgraphGenerator):
    """Generates the subgraph that models the relationship between complex metrics.

    * The successors of simple-metric nodes are metric time nodes / local-model nodes.
    * The successors of complex-metric nodes are other metric nodes (complex or simple).
    """

    @override
    def __init__(self, manifest_object_lookup: ManifestObjectLookup) -> None:
        super().__init__(manifest_object_lookup)
        self._verbose_debug_logs = False

        # Maps the metric name to the corresponding metric node that was generated.
        self._metric_name_to_node: dict[str, MetricNode] = {}
        self._empty_edge_labels: FrozenOrderedSet[MetricFlowGraphLabel] = FrozenOrderedSet()
        self._empty_recipe_step = AttributeRecipeStep()

        common_cumulative_metric_labels = FrozenOrderedSet(
            (
                CumulativeMetricLabel.get_instance(),
                DenyDatePartLabel.get_instance(),
            )
        )

        # Maps the special cases to the labels that should be associated with the edge that connects a metric node
        # to successor nodes.
        self._special_case_to_successor_edge_label = {
            _SpecialCase.CONVERSION_INPUT_METRIC: FrozenOrderedSet((DenyVisibleAttributesLabel.get_instance(),)),
            _SpecialCase.CUMULATIVE_METRIC: common_cumulative_metric_labels,
            _SpecialCase.CUMULATIVE_METRIC_WITH_WINDOW_OR_GRAIN_TO_DATE: common_cumulative_metric_labels.union(
                (DenyEntityKeyQueryResolutionLabel.get_instance(),)
            ),
            _SpecialCase.TIME_OFFSET_DERIVED_METRIC: FrozenOrderedSet(
                (DenyEntityKeyQueryResolutionLabel.get_instance(),)
            ),
        }

    @override
    def add_edges_for_manifest(self, edge_list: list[SemanticGraphEdge]) -> None:
        for metric in self._manifest_object_lookup.get_metrics():
            self._add_edges_for_any_metric(
                metric=metric,
                metric_name_to_node={},
                edge_list=edge_list,
            )

    def _add_edges_for_complex_metric(
        self,
        complex_metric: Metric,
        metric_name_to_node: dict[str, SemanticGraphNode],
        edge_list: list[SemanticGraphEdge],
    ) -> None:
        """Adds the edges from a derived-metric node to the nodes associated with the input metrics."""
        metric_type = complex_metric.type

        input_metric_name_to_labels_for_metric_to_input_metric_edge: dict[
            str, FrozenOrderedSet[MetricFlowGraphLabel]
        ] = {}
        recipe_step = self._empty_recipe_step

        input_metrics = MetricLookup.metric_inputs(complex_metric, include_conversion_metric_input=True)
        if metric_type is MetricType.SIMPLE:
            pass
        elif metric_type is MetricType.CUMULATIVE:
            cumulative_type_params = complex_metric.type_params.cumulative_type_params
            if cumulative_type_params is None:
                raise InvalidManifestException(
                    LazyFormat(
                        "Expected `cumulative_type_params` to be set for a cumulative metric",
                        complex_metric=complex_metric,
                    )
                )

            # Cumulative metrics impose special restrictions on the group-by items available, so label those edges
            # appropriately.

            recipe_step = AttributeRecipeStep(set_deny_date_part=True)
            if cumulative_type_params.window is not None or cumulative_type_params.grain_to_date is not None:
                edge_labels = self._special_case_to_successor_edge_label[
                    _SpecialCase.CUMULATIVE_METRIC_WITH_WINDOW_OR_GRAIN_TO_DATE
                ]
            else:
                edge_labels = self._special_case_to_successor_edge_label[_SpecialCase.CUMULATIVE_METRIC]

            input_metric_for_cumulative_metric = cumulative_type_params.metric
            if input_metric_for_cumulative_metric is None:
                raise InvalidManifestException(
                    LazyFormat(
                        "Expected `metric` to be set for a cumulative metric",
                        complex_metric=complex_metric,
                    )
                )

            input_metric_name_to_labels_for_metric_to_input_metric_edge[
                input_metric_for_cumulative_metric.name
            ] = edge_labels
        elif metric_type is MetricType.RATIO:
            pass
        elif metric_type is MetricType.CONVERSION:
            conversion_type_params = complex_metric.type_params.conversion_type_params
            if conversion_type_params is None:
                raise InvalidManifestException(
                    LazyFormat(
                        "Expected `conversion_type_params` to be set for a conversion metric",
                        complex_metric=complex_metric,
                    )
                )

            conversion_metric = conversion_type_params.conversion_metric
            if conversion_metric is None:
                raise InvalidManifestException(
                    LazyFormat(
                        "Expected `conversion_metric` to be set for a conversion metric", complex_metric=complex_metric
                    )
                )

            input_metric_name_to_labels_for_metric_to_input_metric_edge[
                conversion_metric.name
            ] = self._special_case_to_successor_edge_label[_SpecialCase.CONVERSION_INPUT_METRIC]

        elif metric_type is MetricType.DERIVED:
            pass
        else:
            assert_values_exhausted(metric_type)

        if len(input_metrics) == 0:
            raise RuntimeError(
                LazyFormat(
                    "This method should have been called with a metric that has input metrics",
                    complex_metric=complex_metric,
                    parent_input_metrics=input_metrics,
                )
            )

        complex_metric_node = ComplexMetricNode.get_instance(complex_metric.name)
        additional_edge_labels = self._empty_edge_labels

        for input_metric in input_metrics:
            # Add labels for time-offset derived metrics as that is a special case when resolving the associated
            # group-by items.
            if input_metric.offset_window is not None or input_metric.offset_to_grain is not None:
                additional_edge_labels = self._special_case_to_successor_edge_label[
                    _SpecialCase.TIME_OFFSET_DERIVED_METRIC
                ]
                break

        for input_metric in input_metrics:
            input_metric_name = input_metric.name
            if input_metric_name not in metric_name_to_node:
                self._add_edges_for_any_metric(
                    metric=self._manifest_object_lookup.get_metric(input_metric_name),
                    metric_name_to_node=metric_name_to_node,
                    edge_list=edge_list,
                )
            input_metric_node = metric_name_to_node[input_metric_name]

            edge_to_add = MetricDefinitionEdge.create(
                tail_node=complex_metric_node,
                head_node=input_metric_node,
                additional_labels=additional_edge_labels.union(
                    FrozenOrderedSet(
                        input_metric_name_to_labels_for_metric_to_input_metric_edge.get(input_metric.name) or ()
                    )
                ),
                recipe_step=recipe_step,
            )

            edge_list.append(edge_to_add)
        metric_name_to_node[complex_metric.name] = complex_metric_node

    def _add_edges_for_any_metric(
        self,
        metric: Metric,
        metric_name_to_node: dict[str, SemanticGraphNode],
        edge_list: list[SemanticGraphEdge],
    ) -> None:
        """Adds edges for any type of metric.

        Args:
            metric: Add edges for this metric.
            metric_name_to_node: dict to update when a metric node is added.
            edge_list:  list to update when an edge is added.
        """
        metric_type = metric.type

        if metric_type is MetricType.SIMPLE:
            metric_name = metric.name
            metric_name_to_node[metric_name] = SimpleMetricNode.get_instance(metric_name)
        elif (
            metric_type is MetricType.RATIO
            or metric_type is MetricType.CUMULATIVE
            or metric_type is MetricType.CONVERSION
            or metric_type is MetricType.DERIVED
        ):
            self._add_edges_for_complex_metric(
                complex_metric=metric,
                metric_name_to_node=metric_name_to_node,
                edge_list=edge_list,
            )
        else:
            assert_values_exhausted(metric_type)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Sequence

from dbt_semantic_interfaces.protocols.metric import MetricTimeWindow
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class OffsetCustomGranularityNode(DataflowPlanNode, ABC):
    """Offset a custom grain by the requested number of periods.

    Used to build the time spine node when querying a metric with a custom offset window, when the query requires ONLY the same
    grain as is used in the offset window. The node will output offset columns for all custom grain specs requested and a
    non-offset column for the base grain. The base grain is needed to join to the source node.
    """

    offset_window: MetricTimeWindow
    required_time_spine_specs: Sequence[TimeDimensionSpec]
    time_spine_node: DataflowPlanNode

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        if self.offset_window.is_standard_granularity:
            raise ValueError(
                LazyFormat(
                    "OffsetBaseGrainByCustomGrainNode should only be used for custom grain offset windows.",
                    offset_window=self.offset_window,
                )
            )

    @staticmethod
    def create(  # noqa: D102
        time_spine_node: DataflowPlanNode,
        offset_window: MetricTimeWindow,
        required_time_spine_specs: Sequence[TimeDimensionSpec],
    ) -> OffsetCustomGranularityNode:
        return OffsetCustomGranularityNode(
            parent_nodes=(time_spine_node,),
            time_spine_node=time_spine_node,
            offset_window=offset_window,
            required_time_spine_specs=required_time_spine_specs,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_OFFSET_CUSTOM_GRANULARITY_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_offset_custom_granularity_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Offset Custom Granularity"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (
            DisplayedProperty("offset_window", self.offset_window),
            DisplayedProperty("required_time_spine_specs", self.required_time_spine_specs),
        )

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.offset_window == self.offset_window
            and other_node.required_time_spine_specs == self.required_time_spine_specs
        )

    def with_new_parents(  # noqa: D102
        self, new_parent_nodes: Sequence[DataflowPlanNode]
    ) -> OffsetCustomGranularityNode:
        assert len(new_parent_nodes) == 1
        return OffsetCustomGranularityNode(
            parent_nodes=tuple(new_parent_nodes),
            time_spine_node=new_parent_nodes[0],
            offset_window=self.offset_window,
            required_time_spine_specs=self.required_time_spine_specs,
        )



from __future__ import annotations

import typing
from dataclasses import dataclass
from typing import Optional, Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import override

from metricflow.sql.sql_plan import SqlPlanNode, SqlPlanNodeVisitor, SqlSelectColumn

if typing.TYPE_CHECKING:
    from metricflow.sql.sql_cte_node import SqlCteAliasMapping
    from metricflow.sql.sql_select_node import SqlSelectStatementNode


@dataclass(frozen=True, eq=False)
class SqlTableNode(SqlPlanNode):
    """An SQL table that can go in the FROM clause or the JOIN clause."""

    sql_table: SqlTable

    @staticmethod
    def create(sql_table: SqlTable) -> SqlTableNode:  # noqa: D102
        return SqlTableNode(
            parent_nodes=(),
            sql_table=sql_table,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_PLAN_TABLE_FROM_CLAUSE_ID_PREFIX

    @property
    def description(self) -> str:  # noqa: D102
        return f"Read from {self.sql_table.sql}"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (DisplayedProperty("table_id", self.sql_table.sql),)

    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_table_node(self)

    @property
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:  # noqa: D102
        return None

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        if self.sql_table.schema_name is None:
            cte_node = cte_source_mapping.get_cte_node_for_alias(self.sql_table.table_name)
            if cte_node is not None:
                return cte_node.nearest_select_columns(cte_source_mapping)
        return None

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return self

    @override
    def copy(self) -> SqlTableNode:
        return SqlTableNode.create(
            sql_table=self.sql_table,
        )



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass

from typing_extensions import Optional, override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.formatting.dot_attributes import DotGraphAttributeSet
from metricflow_semantics.experimental.mf_graph.graph_id import MetricFlowGraphId, SequentialGraphId
from metricflow_semantics.experimental.mf_graph.mf_graph import (
    MetricFlowGraph,
    MetricFlowGraphEdge,
    MetricFlowGraphNode,
)
from metricflow_semantics.experimental.mf_graph.mutable_graph import MutableGraph
from metricflow_semantics.experimental.ordered_set import MutableOrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStepProvider,
)
from metricflow_semantics.experimental.semantic_graph.sg_node_grouping import SemanticGraphNodeTypedCollection
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext

logger = logging.getLogger(__name__)


class SemanticGraphNode(MetricFlowGraphNode, AttributeRecipeStepProvider, MetricFlowPrettyFormattable, ABC):
    """A node in the semantic graph.

    At the top level, the semantic graph has two types of nodes: entity nodes and attribute nodes. Entity nodes have
    other entity nodes or attribute nodes as successors while attribute nodes do not have any successors (leaf nodes).

    The entity nodes in the semantic graph are not directly mapped from entities configured in the semantic manifest.
    The one that corresponds to an entity in the manifest are instances of `ConfiguredEntityNode`. However, there
    are other entity nodes that model other relationships. For example, `metric_time` is represented by the
    metric-time entity node, and there is a time-entity node that relates time dimensions and metric time to the
    various time grains (the time grains (e.g. `day`, `year`) map to attribute nodes).

    Please see the subclasses for more details.
    """

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        return self.node_descriptor.node_name

    @property
    @override
    def displayed_properties(self) -> AnyLengthTuple[DisplayedProperty]:
        properties: list[DisplayedProperty] = list(super().displayed_properties)
        if self.recipe_step_to_append is not None:
            properties.extend(self.recipe_step_to_append.displayed_properties)
        properties.extend(DisplayedProperty("label", label) for label in self.labels)
        return tuple(properties)

    @property
    def dunder_name_element(self) -> Optional[str]:
        """This node's associated dunder-name element in the query interface.

        Not all nodes in the semantic graph have a direct relationship to the dunder name, but many do. e.g. the
        metric time entity corresponds to the 'metric_time' dunder-name element.
        """
        return self.recipe_step_to_append.add_dunder_name_element

    @abstractmethod
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        """Add this node to the given typed collection."""
        raise NotImplementedError


@fast_frozen_dataclass(order=False)
class SemanticGraphEdge(MetricFlowGraphEdge[SemanticGraphNode], AttributeRecipeStepProvider, ABC):
    """An edge in the semantic graph.

    Currently, the edges in the semantic graph represent entity relationships and also describe how a related attribute
    can be computed from the relationship (see `AttributeRecipe`).

    For example, if a semantic model contains the `listings` foreign entity and another semantic model contains the
    `listings` primary entity, there would be a path from the node representing the first semantic model to the second
    semantic model (via a configured-entity node for `listings`). The path's edges includes a recipe step that
    describes a join between the two semantic models.
    """

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        formatter = format_context.formatter
        return formatter.pretty_format_object_by_parts(
            class_name=self.__class__.__name__,
            field_mapping={
                "tail_node": self.tail_node,
                "head_node": self.head_node,
            },
        )

    @property
    @override
    def displayed_properties(self) -> AnyLengthTuple[DisplayedProperty]:
        properties: list[DisplayedProperty] = list(super().displayed_properties)
        if self.recipe_step_to_append is not None:
            properties.extend(self.recipe_step_to_append.displayed_properties)

        return tuple(properties)


class SemanticGraph(MetricFlowGraph[SemanticGraphNode, SemanticGraphEdge], ABC):
    """A read-only interface for a semantic graph based on `MetricFlowGraph`.

    Also see `SemanticGraphNode` and `SemanticGraphEdge`.

    The semantic graph helps to model entity relationships that are defined in the semantic manifest and encodes
    context on how attributes for entities can be computed (e.g. by joining semantic models).

    Currently, the edges in the graph are oriented so that a path from a metric node to an attribute node describes
    the query required to compute that metric. However, additional edge types can be added to better model associative
    entity relationships.

    Note: Some changes obviated the inverse graph, and it needs to be either refined or removed.
    """

    @override
    def as_dot_graph(self, include_graphical_attributes: bool) -> DotGraphAttributeSet:
        return (
            super()
            .as_dot_graph(include_graphical_attributes=include_graphical_attributes)
            .with_attributes(
                dot_kwargs={
                    "rankdir": "LR",
                }
                if include_graphical_attributes
                else {}
            )
        )


@dataclass
class MutableSemanticGraph(MutableGraph[SemanticGraphNode, SemanticGraphEdge], SemanticGraph):
    """A mutable implementation of `SemanticGraph` for building graphs."""

    @classmethod
    def create(cls, graph_id: Optional[MetricFlowGraphId] = None) -> MutableSemanticGraph:  # noqa: D102
        return MutableSemanticGraph(
            _graph_id=graph_id or SequentialGraphId.create(),
            _nodes=MutableOrderedSet(),
            _edges=MutableOrderedSet(),
            _tail_node_to_edges=defaultdict(MutableOrderedSet),
            _head_node_to_edges=defaultdict(MutableOrderedSet),
            _label_to_nodes=defaultdict(MutableOrderedSet),
            _label_to_edges=defaultdict(MutableOrderedSet),
            _node_to_predecessor_nodes=defaultdict(MutableOrderedSet),
            _node_to_successor_nodes=defaultdict(MutableOrderedSet),
        )

    @override
    def intersection(self, other: MetricFlowGraph[SemanticGraphNode, SemanticGraphEdge]) -> MutableSemanticGraph:
        intersection_graph = MutableSemanticGraph.create(graph_id=self.graph_id)
        self.add_edges(self._intersect_edges(other))
        return intersection_graph

    @override
    def inverse(self) -> MutableSemanticGraph:
        inverse_graph = MutableSemanticGraph.create()
        for edge in self.edges:
            inverse_graph.add_edge(edge.inverse)
        return inverse_graph

    @override
    def as_sorted(self) -> MutableSemanticGraph:
        updated_graph = MutableSemanticGraph.create(graph_id=self.graph_id)
        for node in sorted(self._nodes):
            updated_graph.add_node(node)

        for edge in sorted(self._edges):
            updated_graph.add_edge(edge)

        return updated_graph



from __future__ import annotations

import logging
from collections import defaultdict
from functools import cached_property
from typing import Iterable, Mapping, Optional, Sequence

from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilterIntersection
from dbt_semantic_interfaces.protocols import Metric, SemanticModel
from dbt_semantic_interfaces.type_enums import DimensionType, MetricType, TimeGranularity
from typing_extensions import override

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.dsi.model_object_lookup import ModelObjectLookup
from metricflow_semantics.experimental.metricflow_exception import InvalidManifestException, MetricFlowInternalError
from metricflow_semantics.mf_logging.attribute_pretty_format import AttributeMapping
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.simple_metric_input import (
    SimpleMetricInput,
    SimpleMetricInputAggregation,
    SimpleMetricInputNonAdditiveDimension,
)

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class SimpleMetricAggregationConfiguration:
    """Key that is used to group the simple metrics in a semantic model by the associated aggregation time dimension."""

    time_dimension_name: str
    time_grain: TimeGranularity


class SimpleMetricModelObjectLookup(ModelObjectLookup):
    """A lookup for models associated with simple metrics.

    A separate lookup class helps to break out the lookup classes and provide better typing (fewer `None` cases).
    """

    def __init__(  # noqa: D107
        self, semantic_model: SemanticModel, simple_metrics: Optional[Sequence[Metric]] = None
    ) -> None:
        super().__init__(semantic_model)
        model_name = semantic_model.name
        # Sanity checks.
        if not simple_metrics:
            raise MetricFlowInternalError(
                LazyFormat(
                    "Can't initialize with empty `simple_metrics`",
                    simple_metrics=simple_metrics,
                )
            )
        for metric in simple_metrics:
            if metric.type is not MetricType.SIMPLE:
                raise MetricFlowInternalError(
                    LazyFormat("Can't initialize with a metric that is not a simple metric", metric=metric)
                )
            metric_aggregation_params = metric.type_params.metric_aggregation_params
            if metric_aggregation_params is None or metric_aggregation_params.semantic_model != model_name:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "Can't initialize with a metric that is not associated with this semantic model",
                        metric=metric,
                        model_name=model_name,
                    )
                )
        self._simple_metrics = simple_metrics

    @cached_property
    def _simple_metric_inputs_from_metrics(self) -> Sequence[SimpleMetricInput]:
        simple_metric_inputs: list[SimpleMetricInput] = []
        for metric in self._simple_metrics:
            metric_type_params = metric.type_params
            metric_aggregation_params = metric.type_params.metric_aggregation_params
            assert metric_aggregation_params is not None, "`metric_aggregation_params` should be set for all metrics"

            agg_time_dimension_name = metric_aggregation_params.agg_time_dimension or (
                self._semantic_model.defaults.agg_time_dimension if self._semantic_model.defaults is not None else None
            )
            agg_time_dimension_grain = (
                self._time_dimension_name_to_grain.get(agg_time_dimension_name)
                if agg_time_dimension_name is not None
                else None
            )
            if agg_time_dimension_name is None or agg_time_dimension_grain is None:
                raise InvalidManifestException(
                    LazyFormat(
                        "Invalid aggregation time dimension configuration",
                        metric=metric,
                        semantic_model=self._semantic_model,
                    )
                )

            simple_metric_input = SimpleMetricInput(
                name=metric.name,
                agg=metric_aggregation_params.agg,
                expr=metric_type_params.expr or metric.name,
                agg_params=SimpleMetricInputAggregation.create_from_pydantic(metric_aggregation_params.agg_params),
                join_to_timespine=metric.type_params.join_to_timespine,
                fill_nulls_with=metric.type_params.fill_nulls_with,
                non_additive_dimension=SimpleMetricInputNonAdditiveDimension.create_from_non_additive_dimension(
                    metric_aggregation_params.non_additive_dimension
                ),
                agg_time_dimension_name=agg_time_dimension_name,
                agg_time_dimension_grain=agg_time_dimension_grain,
                model_id=self.model_id,
                filter=metric.filter or PydanticWhereFilterIntersection(where_filters=[]),
            )
            simple_metric_inputs.append(simple_metric_input)

        return simple_metric_inputs

    @cached_property
    def aggregation_time_dimension_name_to_simple_metric_inputs(self) -> Mapping[str, Iterable[SimpleMetricInput]]:
        """Mapping from the name of the aggregation time dimension to the `SimpleMetricInputs` for this model."""
        current_aggregation_time_dimension_name_to_simple_metric_inputs: dict[
            str, list[SimpleMetricInput]
        ] = defaultdict(list)

        for simple_metric_input in self._simple_metric_inputs_from_metrics:
            current_aggregation_time_dimension_name_to_simple_metric_inputs[
                simple_metric_input.agg_time_dimension_name
            ].append(simple_metric_input)

        return current_aggregation_time_dimension_name_to_simple_metric_inputs

    @cached_property
    def aggregation_configuration_to_simple_metric_inputs(
        self,
    ) -> Mapping[SimpleMetricAggregationConfiguration, Iterable[SimpleMetricInput]]:
        """Mapping from the aggregation configuration to the `SimpleMetricInputs` for this model."""
        current_aggregation_configuration_to_simple_metric_inputs: defaultdict[
            SimpleMetricAggregationConfiguration, list[SimpleMetricInput]
        ] = defaultdict(list)

        for simple_metric_input in self._simple_metric_inputs_from_metrics:
            aggregation_configuration = SimpleMetricAggregationConfiguration(
                time_dimension_name=simple_metric_input.agg_time_dimension_name,
                time_grain=simple_metric_input.agg_time_dimension_grain,
            )

            current_aggregation_configuration_to_simple_metric_inputs[aggregation_configuration].append(
                simple_metric_input
            )

        return current_aggregation_configuration_to_simple_metric_inputs

    @cached_property
    def _time_dimension_name_to_grain(self) -> Mapping[str, TimeGranularity]:
        return {
            dimension.name: dimension.type_params.time_granularity
            for dimension in self._semantic_model.dimensions
            if (dimension.type is DimensionType.TIME and dimension.type_params is not None)
        }

    @cached_property
    @override
    def _attribute_mapping(self) -> AttributeMapping:
        return dict(
            **super()._attribute_mapping,
            **{
                "aggregation_configuration_to_simple_metric_inputs": {
                    configuration: [simple_metric_input.name for simple_metric_input in simple_metric_inputs]
                    for configuration, simple_metric_inputs in self.aggregation_configuration_to_simple_metric_inputs.items()
                }
            },
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Sequence

from dbt_semantic_interfaces.protocols import Dimension, SemanticModel
from dbt_semantic_interfaces.references import DimensionReference
from dbt_semantic_interfaces.type_enums import DimensionType

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.semantic_model_helper import SemanticModelHelper
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName


@dataclass(frozen=True)
class DimensionInvariant:
    """For a given manifest, all defined dimensions with the same name should have these same properties."""

    dimension_type: DimensionType
    is_partition: bool


class DimensionLookup:
    """Looks up properties related to dimensions."""

    def __init__(self, semantic_models: Sequence[SemanticModel]) -> None:  # noqa: D107
        self._dimension_reference_to_invariant: Dict[DimensionReference, DimensionInvariant] = {}
        self.dimensions_by_dunder_name: Dict[str, Dimension] = {}
        for semantic_model in semantic_models:
            for dimension in semantic_model.dimensions:
                invariant = DimensionInvariant(
                    dimension_type=dimension.type,
                    is_partition=dimension.is_partition,
                )
                dimension_reference = dimension.reference
                existing_invariant = self._dimension_reference_to_invariant.get(dimension_reference)
                if existing_invariant is not None and existing_invariant != invariant:
                    raise ValueError(
                        LazyFormat(
                            "Dimensions with the same name have been defined with conflicting values that "
                            "should have been the same in a given semantic manifest. This should have been caught "
                            "during validation.",
                            dimension_reference=dimension_reference,
                            existing_invariant=existing_invariant,
                            conflicting_invariant=invariant,
                            semantic_model_reference=semantic_model.reference,
                        )
                    )

                self._dimension_reference_to_invariant[dimension_reference] = invariant
                primary_entity = SemanticModelHelper.resolved_primary_entity(semantic_model)
                structured_name = StructuredLinkableSpecName(
                    entity_link_names=(primary_entity.element_name,), element_name=dimension.name
                )
                self.dimensions_by_dunder_name[structured_name.dunder_name] = dimension

    def get_invariant(self, dimension_reference: DimensionReference) -> DimensionInvariant:
        """Get invariants for the given dimension in the semantic manifest."""
        # dimension_reference might be a TimeDimensionReference, so change types.
        dimension_reference = DimensionReference(element_name=dimension_reference.element_name)
        invariant = self._dimension_reference_to_invariant[dimension_reference]
        if invariant is None:
            raise ValueError(
                LazyFormat(
                    "Unknown dimension reference",
                    dimension_reference=dimension_reference,
                    known_dimension_references=list(self._dimension_reference_to_invariant.keys()),
                )
            )

        return invariant



from __future__ import annotations

import logging

from _pytest.fixtures import FixtureRequest
from dbt_semantic_interfaces.implementations.semantic_manifest import PydanticSemanticManifest
from dbt_semantic_interfaces.transformations.semantic_manifest_transformer import PydanticSemanticManifestTransformer
from dbt_semantic_interfaces.validations.semantic_manifest_validator import SemanticManifestValidator

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.test_helpers.config_helpers import MetricFlowTestConfiguration
from metricflow_semantics.test_helpers.snapshot_helpers import assert_object_snapshot_equal
from metricflow_semantics.test_helpers.synthetic_manifest.semantic_manifest_generator import SyntheticManifestGenerator
from metricflow_semantics.test_helpers.synthetic_manifest.synthetic_manifest_parameter_set import (
    SyntheticManifestParameterSet,
)

logger = logging.getLogger(__name__)


def test_manifest_generator(  # noqa: D103
    request: FixtureRequest,
    mf_test_configuration: MetricFlowTestConfiguration,
) -> None:
    parameter_set = SyntheticManifestParameterSet(
        simple_metric_semantic_model_count=2,
        simple_metrics_per_semantic_model=2,
        dimension_semantic_model_count=2,
        categorical_dimensions_per_semantic_model=2,
        max_metric_depth=2,
        max_metric_width=2,
        saved_query_count=2,
        metrics_per_saved_query=2,
        categorical_dimensions_per_saved_query=2,
    )
    generator = SyntheticManifestGenerator(parameter_set)
    manifest = generator.generate_manifest()
    assert_object_snapshot_equal(
        request=request,
        snapshot_configuration=mf_test_configuration,
        obj=manifest,
    )

    manifest = PydanticSemanticManifestTransformer.transform(manifest)
    validator = SemanticManifestValidator[PydanticSemanticManifest]()
    validation_result = validator.validate_semantic_manifest(manifest)
    logger.debug(LazyFormat("Generated manifest", manifest=manifest))

    assert not validation_result.has_blocking_issues, LazyFormat(
        "Found validation issues with the generated manifest",
        validation_result=validation_result,
        manifest=manifest,
    ).evaluated_value



from __future__ import annotations

from functools import cached_property

from metricflow_semantics.test_helpers.synthetic_manifest.synthetic_manifest_parameter_set import (
    SyntheticManifestParameterSet,
)


class CategoricalDimensionGenerator:
    """Helps generate the categorical dimensions in the semantic manifest.

    The index for the dimension refers to the index when all unique dimensions in the semantic manifest are enumerated.
    """

    def __init__(self, parameter_set: SyntheticManifestParameterSet) -> None:  # noqa: D107
        self._parameter_set = parameter_set

    def get_dimension_name(self, dimension_index: int) -> str:  # noqa: D102
        """Return the name of the dimension for the given index."""
        return f"dimension_{dimension_index:03}"

    @cached_property
    def unique_dimension_count(self) -> int:  # noqa: D102
        return (
            self._parameter_set.categorical_dimensions_per_semantic_model
            * self._parameter_set.dimension_semantic_model_count
        )

    def get_next_wrapped_index(self, dimension_index: int) -> int:
        """Return the next valid dimension index, wrapping back to 0 if it reaches the last index."""
        if dimension_index < 0:
            raise ValueError(f"{dimension_index=} should be > 0")

        if dimension_index >= self.unique_dimension_count:
            raise ValueError(f"{dimension_index=} should be < {self.unique_dimension_count}")

        return (dimension_index + 1) % self.unique_dimension_count



from __future__ import annotations

import logging
from functools import cached_property
from typing import Dict, Final, Iterable, Sequence

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.naming.keywords import METRIC_TIME_ELEMENT_NAME
from dbt_semantic_interfaces.protocols import MetricInput
from dbt_semantic_interfaces.protocols.metric import Metric, MetricType
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.references import MetricReference, SemanticModelReference
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow_semantics.errors.error_classes import (
    DuplicateMetricError,
    MetricNotFoundError,
    UnknownMetricError,
)
from metricflow_semantics.experimental.cache.mf_cache import ResultCache
from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.metricflow_exception import InvalidManifestException, MetricFlowInternalError
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet, OrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.annotated_spec_linkable_element_set import (
    GroupByItemSet,
)
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantics.element_filter import GroupByItemSetFilter
from metricflow_semantics.model.semantics.linkable_element_set_base import BaseGroupByItemSet
from metricflow_semantics.model.semantics.linkable_spec_resolver import (
    GroupByItemSetResolver,
)
from metricflow_semantics.specs.spec_set import group_specs_by_type
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec

logger = logging.getLogger(__name__)


DEFAULT_COMMON_SET_FILTER: Final[GroupByItemSetFilter] = GroupByItemSetFilter.create(
    any_properties_denylist=(GroupByItemProperty.METRIC,)
)


class MetricLookup:
    """Tracks semantic information for metrics by linking them to semantic models."""

    def __init__(
        self,
        semantic_manifest: SemanticManifest,
        group_by_item_set_resolver: GroupByItemSetResolver,
        manifest_object_lookup: ManifestObjectLookup,
    ) -> None:
        """Initializer.

        Args:
            semantic_manifest: used to fetch and load the metrics and initialize the linkable spec resolver
        """
        self._metrics: Dict[MetricReference, Metric] = {}
        self._manifest_object_lookup = manifest_object_lookup

        for metric in semantic_manifest.metrics:
            metric_reference = MetricReference(element_name=metric.name)
            if metric_reference in self._metrics:
                raise DuplicateMetricError(
                    LazyFormat(
                        "A duplicate metric was found in the manifest",
                        conflicting_metrics=[self._metrics[metric_reference], metric],
                    )
                )
            self._metrics[metric_reference] = metric

        self._group_by_item_set_resolver = group_by_item_set_resolver

        self._result_cache_for_get_min_queryable_time_granularity: ResultCache[
            MetricReference, TimeGranularity
        ] = ResultCache()
        self._result_cache_for_aggregation_time_dimension_specs: ResultCache[
            str, FrozenOrderedSet[TimeDimensionSpec]
        ] = ResultCache()
        self._result_cache_for_derived_from_semantic_models: ResultCache[
            MetricReference, FrozenOrderedSet[SemanticModelReference]
        ] = ResultCache()

    def get_group_by_items_for_distinct_values_query(
        self, set_filter: GroupByItemSetFilter = GroupByItemSetFilter.create()
    ) -> BaseGroupByItemSet:
        """Return the reachable linkable elements for a dimension values query with no metrics."""
        return self._group_by_item_set_resolver.get_set_for_distinct_values_query(set_filter)

    def get_derived_from_semantic_models(self, metric_reference: MetricReference) -> OrderedSet[SemanticModelReference]:
        """Return the semantic models that the given metric is derived from.

        i.e. the set of semantic models where constituent simple metrics are located.
        """
        result_cache = self._result_cache_for_derived_from_semantic_models
        result = result_cache.get(metric_reference)
        if result:
            return result.value

        metric = self.get_metric(metric_reference)
        metric_inputs = MetricLookup.metric_inputs(metric, include_conversion_metric_input=True)

        model_references: MutableOrderedSet[SemanticModelReference] = MutableOrderedSet()
        if len(metric_inputs) == 0:
            metric_aggregation_params = metric.type_params.metric_aggregation_params
            if metric_aggregation_params is None:
                raise MetricFlowInternalError(
                    LazyFormat("Expected `metric_aggregation_params` to be set", metric=metric)
                )
            model_references.add(SemanticModelReference(metric_aggregation_params.semantic_model))
        else:
            model_references.update(
                *(
                    self.get_derived_from_semantic_models(MetricReference(metric_input.name))
                    for metric_input in metric_inputs
                )
            )

        return self._result_cache_for_derived_from_semantic_models.set_and_get(
            metric_reference, model_references.as_frozen()
        )

    def get_common_group_by_items(
        self,
        metric_references: Iterable[MetricReference] = (),
        set_filter: GroupByItemSetFilter = DEFAULT_COMMON_SET_FILTER,
        joins_disallowed: bool = False,
    ) -> BaseGroupByItemSet:
        """Gets the set of the valid group-by items common to all inputs."""
        if set_filter.element_name_allowlist is None:
            return self._group_by_item_set_resolver.get_common_set(
                metric_references=metric_references,
                set_filter=set_filter,
                joins_disallowed=joins_disallowed,
            )

        # If the filter specifies element names, make the call to the resolver without element names to get better
        # cache hit rates.
        filter_without_element_name_condition = set_filter.without_element_name_allowlist()

        result_superset = self._group_by_item_set_resolver.get_common_set(
            metric_references=metric_references,
            set_filter=filter_without_element_name_condition,
            joins_disallowed=joins_disallowed,
        )

        return GroupByItemSet(
            annotated_specs=tuple(
                annotated_spec
                for annotated_spec in result_superset.annotated_specs
                if annotated_spec.element_name in set_filter.element_name_allowlist
            )
        )

    def get_metrics(self, metric_references: Iterable[MetricReference]) -> Sequence[Metric]:  # noqa: D102
        return tuple(self.get_metric(metric_reference) for metric_reference in metric_references)

    @cached_property
    def metric_references(self) -> OrderedSet[MetricReference]:  # noqa: D102
        return FrozenOrderedSet(sorted(self._metrics.keys()))

    def get_metric(self, metric_reference: MetricReference) -> Metric:  # noqa: D102
        metric = self._metrics.get(metric_reference)
        if metric is None:
            raise MetricNotFoundError(LazyFormat("The given metric is not known", metric_reference=metric_reference))
        return metric

    @staticmethod
    def metric_inputs(metric: Metric, include_conversion_metric_input: bool) -> Sequence[MetricInput]:
        """Returns the metric inputs for the given metric."""
        metric_type = metric.type
        metric_inputs: list[MetricInput] = []

        if metric_type is MetricType.SIMPLE:
            pass
        elif metric_type is MetricType.CUMULATIVE:
            cumulative_type_params = metric.type_params.cumulative_type_params
            if cumulative_type_params is None:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "Expected `cumulative_type_params` to be set for a cumulative metric",
                        complex_metric=metric,
                    )
                )

            input_metric_for_cumulative_metric = cumulative_type_params.metric
            if input_metric_for_cumulative_metric is None:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "Expected `metric` to be set for a cumulative metric",
                        complex_metric=metric,
                    )
                )

            metric_inputs.append(input_metric_for_cumulative_metric)
        elif metric_type is MetricType.RATIO:
            numerator = metric.type_params.numerator
            if numerator is None:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "Expected `numerator` to be set for a ratio metric",
                        complex_metric=metric,
                    )
                )
            metric_inputs.append(numerator)

            denominator = metric.type_params.denominator
            if denominator is None:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "Expected `denominator` to be set for a ratio metric",
                        complex_metric=metric,
                    )
                )
            metric_inputs.append(denominator)
        elif metric_type is MetricType.CONVERSION:
            conversion_type_params = metric.type_params.conversion_type_params
            if conversion_type_params is None:
                raise MetricFlowInternalError(
                    LazyFormat(
                        "Expected `conversion_type_params` to be set for a conversion metric",
                        complex_metric=metric,
                    )
                )
            base_metric = conversion_type_params.base_metric
            if base_metric is None:
                raise MetricFlowInternalError(
                    LazyFormat("Expected `base_metric` to be set for a conversion metric", complex_metric=metric)
                )
            metric_inputs.append(base_metric)

            if include_conversion_metric_input:
                conversion_metric = conversion_type_params.conversion_metric
                if conversion_metric is None:
                    raise MetricFlowInternalError(
                        LazyFormat(
                            "Expected `conversion_metric` to be set for a conversion metric", complex_metric=metric
                        )
                    )
                metric_inputs.append(conversion_metric)

        elif metric_type is MetricType.DERIVED:
            metrics = metric.type_params.metrics
            if not metrics:
                raise MetricFlowInternalError(
                    LazyFormat("Expected `metrics` to be set for a derived metric", derived_metric=metric)
                )
            metric_inputs.extend(metrics)
        else:
            assert_values_exhausted(metric_type)

        return metric_inputs

    def get_aggregation_time_dimension_specs(self, metric_reference: MetricReference) -> OrderedSet[TimeDimensionSpec]:
        """Return the time dimension specs that are used for aggregating the simple metric."""
        metric_name = metric_reference.element_name
        cache_key = metric_name
        result = self._result_cache_for_aggregation_time_dimension_specs.get(cache_key)
        if result:
            return result.value

        metric_inputs = MetricLookup.metric_inputs(
            self.get_metric(metric_reference), include_conversion_metric_input=False
        )
        if len(metric_inputs) > 0:
            intersection_result = MutableOrderedSet(
                self.get_aggregation_time_dimension_specs(MetricReference(metric_inputs[0].name))
            )
            for metric_input in metric_inputs[1:]:
                intersection_result.intersection_update(
                    self.get_aggregation_time_dimension_specs(MetricReference(metric_input.name))
                )
        else:
            simple_metric_input = self._manifest_object_lookup.simple_metric_name_to_input.get(metric_name)
            if simple_metric_input is None:
                raise ValueError(
                    LazyFormat(
                        "Unable to find a simple metric with the given name",
                        metric_name=metric_name,
                    )
                )

            group_by_item_set = self.get_common_group_by_items(
                metric_references=(metric_reference,),
                set_filter=GroupByItemSetFilter.create(
                    element_name_allowlist=(simple_metric_input.agg_time_dimension_name, METRIC_TIME_ELEMENT_NAME)
                ),
                joins_disallowed=True,
            )
            spec_set = group_specs_by_type(group_by_item_set.specs)
            intersection_result = MutableOrderedSet(spec_set.time_dimension_specs)

        return self._result_cache_for_aggregation_time_dimension_specs.set_and_get(
            cache_key, intersection_result.as_frozen()
        )

    def get_min_queryable_time_granularity(self, metric_reference: MetricReference) -> TimeGranularity:
        """The minimum grain that can be queried with this metric.

        Maps to the largest granularity defined for any of the metric's agg_time_dimensions.
        """
        cache = self._result_cache_for_get_min_queryable_time_granularity
        cache_key = metric_reference
        result = cache.get(cache_key)
        if result:
            return result.value

        metric = self.get_metric(metric_reference)
        metric_type = metric.type

        agg_time_dimension_grains: set[TimeGranularity] = set()

        if metric_type is MetricType.SIMPLE:
            metric_name = metric_reference.element_name
            simple_metric_input = self._manifest_object_lookup.simple_metric_name_to_input.get(metric_name)
            if simple_metric_input is None:
                raise UnknownMetricError((metric_name,))
            agg_time_dimension_grains.add(simple_metric_input.agg_time_dimension_grain)
        elif (
            metric_type is MetricType.CONVERSION
            or metric_type is MetricType.DERIVED
            or metric_type is MetricType.RATIO
            or metric_type is MetricType.CUMULATIVE
        ):
            metric_inputs = MetricLookup.metric_inputs(metric, include_conversion_metric_input=True)
            if not metric_inputs:
                raise InvalidManifestException(
                    LazyFormat("Expected `metrics` to be non-empty for a non-simple metric", metric=metric)
                )
            agg_time_dimension_grains.update(
                (
                    self.get_min_queryable_time_granularity(MetricReference(metric_input.name))
                    for metric_input in metric_inputs
                )
            )
        else:
            assert_values_exhausted(metric_type)

        if len(agg_time_dimension_grains) == 0:
            raise RuntimeError(
                LazyFormat("Unable to resolve an aggregation-time-dimension grain for the given metric", metric=metric)
            )

        return cache.set_and_get(
            key=cache_key, value=max(agg_time_dimension_grains, key=lambda time_grain: time_grain.to_int())
        )



from __future__ import annotations

from abc import ABC

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.dag.mf_dag import DisplayedProperty


class HasDisplayedProperty(ABC):
    """Mixin for classes that have properties that should be displayed in visualizations."""

    @property
    def displayed_properties(self) -> AnyLengthTuple[DisplayedProperty]:
        """Return the properties that should be displayed in visualization."""
        return ()


class MetricFlowGraphElement(HasDisplayedProperty, ABC):
    """An element in a graph (e.g. node)."""

    pass



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Generic, Optional

from typing_extensions import override

from metricflow_semantics.experimental.mf_graph.mutable_graph import EdgeT, NodeT
from metricflow_semantics.experimental.mf_graph.path_finding.graph_path import MutablePathT

logger = logging.getLogger(__name__)


class WeightFunction(Generic[NodeT, EdgeT, MutablePathT], ABC):
    """The weight function that is used to limit paths during graph traversal."""

    @abstractmethod
    def incremental_weight(self, path_to_node: MutablePathT, next_edge: EdgeT) -> Optional[int]:
        """Return the incremental weight added by adding the given edge to the path."""
        raise NotImplementedError()


class EdgeCountWeightFunction(Generic[NodeT, EdgeT, MutablePathT], WeightFunction[NodeT, EdgeT, MutablePathT]):
    """Translates the number of edges in the path to the weight."""

    @override
    def incremental_weight(self, path_to_node: MutablePathT, next_edge: EdgeT) -> Optional[int]:
        return 1



from __future__ import annotations

import logging
import typing

from typing_extensions import override

from metricflow_semantics.experimental.mf_graph.formatting.graph_formatter import MetricFlowGraphFormatter
from metricflow_semantics.mf_logging.pretty_print import mf_pformat

if typing.TYPE_CHECKING:
    from metricflow_semantics.experimental.mf_graph.mf_graph import MetricFlowGraph

logger = logging.getLogger(__name__)


class PrettyFormatGraphFormatter(MetricFlowGraphFormatter):
    """Formats a graph using `mf_pformat()`."""

    @override
    def format_graph(self, graph: MetricFlowGraph) -> str:
        return mf_pformat(graph)



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence, Tuple

from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class DuplicateOutputColumnIssue(MetricFlowQueryResolutionIssue):
    """Describes when there are duplicate output column names in a query."""

    duplicate_column_names: Tuple[str, ...]

    @staticmethod
    def from_parameters(  # noqa: D102
        duplicate_column_names: Sequence[str],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> DuplicateOutputColumnIssue:
        return DuplicateOutputColumnIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            duplicate_column_names=tuple(duplicate_column_names),
            query_resolution_path=query_resolution_path,
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return f"Query contains duplicate output column names: {self.duplicate_column_names}"

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> DuplicateOutputColumnIssue:
        return DuplicateOutputColumnIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            duplicate_column_names=self.duplicate_column_names,
        )



from __future__ import annotations

import itertools
import logging
import typing
from abc import ABC, abstractmethod

from typing_extensions import override

from metricflow_semantics.helpers.table_helpers import IsolatedTabulateRunner

if typing.TYPE_CHECKING:
    from metricflow_semantics.test_helpers.performance.performance_helpers import SessionReport

logger = logging.getLogger(__name__)


class SessionReportTextFormatter(ABC):
    """Interface for formatting performance profile reports to text (e.g. table, markdown)."""

    @abstractmethod
    def format_report(self, report: SessionReport) -> str:
        """Format a session report according to this formatter."""
        raise NotImplementedError


class TableTextFormatter(SessionReportTextFormatter):
    """Formats a `SessionReport` to a text table for logs."""

    def __init__(self, row_limit: int = 40) -> None:  # noqa: D107
        self._row_limit = row_limit

    def _justify_and_pad_column(
        self, rows: list[dict[str, str]], column_name: str, left_justify: bool = True, pad_character: str = " "
    ) -> None:
        """Pad / justify the columns e.g. so that the numbers place-values line up.

        If `left_justify` is not set, it will be right justified.
        """
        max_column_length: int = max(itertools.chain((len(column_name),), (len(row[column_name]) for row in rows)))
        justify_character = "<" if left_justify else ">"
        for row in rows:
            previous_value = row[column_name]
            row[column_name] = f"{previous_value:{pad_character}{justify_character}{max_column_length}}"

    @override
    def format_report(self, report: SessionReport) -> str:
        rows: list[dict[str, str]] = []
        for function_report in report.functions.values():
            rows.append(
                {
                    "total": f"{function_report.total_time:.3f}",
                    "body": f"{function_report.body_time:.3f}",
                    "function_name": function_report.function_name,
                    "calls": str(function_report.total_calls),
                    "base_calls": str(function_report.base_calls),
                }
            )
        # Justify numerical columns.
        for column_name in ("total", "body", "calls", "base_calls"):
            self._justify_and_pad_column(rows, column_name, left_justify=False)
        # Add middle dot to make it easier to see the associated row values.
        self._justify_and_pad_column(rows, "function_name", pad_character="·")

        rows.sort(reverse=True, key=lambda row: (row["total"], row["body"], row["calls"], row["base_calls"]))

        items = (
            f"Session ID: {report.session_id}",
            IsolatedTabulateRunner.tabulate(
                tabular_data=rows[: self._row_limit],
                headers="keys",
                tablefmt="simple_outline",
            ),
        )
        return "\n".join(items)



from __future__ import annotations

import logging

from metricflow_semantics.experimental.metricflow_exception import MetricFlowInternalError

logger = logging.getLogger(__name__)


class SemanticGraphTraversalError(MetricFlowInternalError):
    """Raised when an unexpected condition is encountered during semantic-graph traversal."""

    pass



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.implementations.elements.dimension import PydanticDimensionTypeParams
from dbt_semantic_interfaces.protocols.dimension import (
    Dimension as SemanticManifestDimension,
)
from dbt_semantic_interfaces.protocols.dimension import (
    DimensionType,
    DimensionTypeParams,
)
from dbt_semantic_interfaces.protocols.entity import Entity as SemanticManifestEntity
from dbt_semantic_interfaces.protocols.export import Export
from dbt_semantic_interfaces.protocols.metadata import Metadata
from dbt_semantic_interfaces.protocols.metric import Metric as SemanticManifestMetric
from dbt_semantic_interfaces.protocols.metric import (
    MetricType,
    MetricTypeParams,
    SemanticLayerElementConfig,
)
from dbt_semantic_interfaces.protocols.saved_query import (
    SavedQuery as SemanticManifestSavedQuery,
)
from dbt_semantic_interfaces.protocols.saved_query import (
    SavedQueryQueryParams,
)
from dbt_semantic_interfaces.protocols.where_filter import WhereFilterIntersection
from dbt_semantic_interfaces.references import EntityReference, SemanticModelReference
from dbt_semantic_interfaces.type_enums.entity_type import EntityType
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.specs.dimension_spec import DimensionSpec


class SearchableElement(ABC):
    """Base class for searchable elements."""

    @property
    @abstractmethod
    def default_search_and_sort_attribute(self) -> str:  # noqa: D102
        """The default attribute to use when filtering elements by a search string."""
        raise NotImplementedError


@dataclass(frozen=True)
class Metric(SearchableElement):
    """Dataclass representation of a Metric."""

    name: str
    description: Optional[str]
    type: MetricType
    type_params: MetricTypeParams
    filter: Optional[WhereFilterIntersection]
    metadata: Optional[Metadata]
    dimensions: List[Dimension]
    label: Optional[str]
    config: Optional[SemanticLayerElementConfig]
    semantic_models: List[SemanticModelReference]

    @classmethod
    def from_pydantic(
        cls,
        pydantic_metric: SemanticManifestMetric,
        dimensions: List[Dimension],
        semantic_models: List[SemanticModelReference],
    ) -> Metric:
        """Build from pydantic Metric object and list of Dimensions."""
        return cls(
            name=pydantic_metric.name,
            description=pydantic_metric.description,
            type=pydantic_metric.type,
            type_params=pydantic_metric.type_params,
            filter=pydantic_metric.filter,
            metadata=pydantic_metric.metadata,
            dimensions=dimensions,
            label=pydantic_metric.label,
            config=pydantic_metric.config,
            semantic_models=semantic_models,
        )

    @property
    def default_search_and_sort_attribute(self) -> str:  # noqa: D102
        return self.name


@dataclass(frozen=True)
class Dimension(SearchableElement):
    """Dataclass representation of a Dimension."""

    name: str
    dunder_name: str
    description: Optional[str]
    type: DimensionType
    entity_links: Tuple[EntityReference, ...]
    type_params: Optional[DimensionTypeParams]
    metadata: Optional[Metadata]
    semantic_model_reference: Optional[SemanticModelReference]
    config: Optional[SemanticLayerElementConfig] = None
    is_partition: bool = False
    expr: Optional[str] = None
    label: Optional[str] = None

    @classmethod
    def from_pydantic(
        cls,
        pydantic_dimension: SemanticManifestDimension,
        entity_links: Tuple[EntityReference, ...],
        semantic_model_reference: SemanticModelReference,
    ) -> Dimension:
        """Build from pydantic Dimension and entity_key."""
        qualified_name = DimensionSpec(element_name=pydantic_dimension.name, entity_links=entity_links).dunder_name
        parsed_type_params: Optional[DimensionTypeParams] = None
        if pydantic_dimension.type_params:
            parsed_type_params = PydanticDimensionTypeParams(
                time_granularity=pydantic_dimension.type_params.time_granularity,
                validity_params=pydantic_dimension.type_params.validity_params,
            )
        return cls(
            name=pydantic_dimension.name,
            dunder_name=qualified_name,
            description=pydantic_dimension.description,
            type=pydantic_dimension.type,
            type_params=parsed_type_params,
            metadata=pydantic_dimension.metadata,
            config=pydantic_dimension.config,
            is_partition=pydantic_dimension.is_partition,
            expr=pydantic_dimension.expr,
            label=pydantic_dimension.label,
            entity_links=entity_links,
            semantic_model_reference=semantic_model_reference,
        )

    @property
    def default_search_and_sort_attribute(self) -> str:  # noqa: D102
        return self.dunder_name

    @property
    def granularity_free_dunder_name(self) -> str:
        """Renders the qualified name without the granularity suffix.

        In the list metrics and list dimensions outputs we want to render the qualified name of the dimension, but
        without including the base granularity for time dimensions. This method is useful in those contexts.

        Note: in most cases you should be using the dunder_name - this is only useful in cases where the
        Dimension set has de-duplicated TimeDimensions such that you never have more than one granularity
        in your set for each TimeDimension.
        """
        return StructuredLinkableSpecName(
            entity_link_names=tuple(e.element_name for e in self.entity_links), element_name=self.name
        ).dunder_name


@dataclass(frozen=True)
class Entity(SearchableElement):
    """Dataclass representation of a Entity."""

    name: str
    description: Optional[str]
    type: EntityType
    semantic_model_reference: SemanticModelReference
    role: Optional[str]
    config: Optional[SemanticLayerElementConfig] = None
    expr: Optional[str] = None

    @classmethod
    def from_pydantic(
        cls, pydantic_entity: SemanticManifestEntity, semantic_model_reference: SemanticModelReference
    ) -> Entity:
        """Build from pydantic Entity."""
        return cls(
            name=pydantic_entity.name,
            description=pydantic_entity.description,
            type=pydantic_entity.type,
            role=pydantic_entity.role,
            config=pydantic_entity.config,
            expr=pydantic_entity.expr,
            semantic_model_reference=semantic_model_reference,
        )

    @property
    def default_search_and_sort_attribute(self) -> str:  # noqa: D102
        return self.name


@dataclass(frozen=True)
class SavedQuery(SearchableElement):
    """Dataclass representation of a SavedQuery."""

    name: str
    description: Optional[str]
    label: Optional[str]
    query_params: SavedQueryQueryParams
    metadata: Optional[Metadata]
    exports: Sequence[Export]
    tags: Sequence[str]

    @classmethod
    def from_pydantic(cls, pydantic_saved_query: SemanticManifestSavedQuery) -> SavedQuery:
        """Build from pydantic SavedQuery object."""
        return cls(
            name=pydantic_saved_query.name,
            description=pydantic_saved_query.description,
            label=pydantic_saved_query.label,
            query_params=pydantic_saved_query.query_params,
            metadata=pydantic_saved_query.metadata,
            exports=pydantic_saved_query.exports,
            tags=pydantic_saved_query.tags,
        )

    @property
    def default_search_and_sort_attribute(self) -> str:  # noqa: D102
        return self.name



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class WriteToResultTableNode(DataflowPlanNode):
    """A node where incoming data gets written to a table.

    Attributes:
        output_sql_table: The table where the computed metrics should be written to.
    """

    output_sql_table: SqlTable

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        output_sql_table: SqlTable,
    ) -> WriteToResultTableNode:
        return WriteToResultTableNode(
            parent_nodes=(parent_node,),
            output_sql_table=output_sql_table,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_WRITE_TO_RESULT_DATA_TABLE_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_write_to_result_table_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Write to Table"""

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return isinstance(other_node, self.__class__) and other_node.output_sql_table == self.output_sql_table

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> WriteToResultTableNode:  # noqa: D102
        return WriteToResultTableNode.create(
            parent_node=new_parent_nodes[0],
            output_sql_table=self.output_sql_table,
        )



from __future__ import annotations

import logging
from abc import ABC
from dataclasses import dataclass
from typing import Generic, Mapping

from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.mf_graph.mutable_graph import NodeT
from metricflow_semantics.experimental.mf_graph.path_finding.traversal_profile import GraphTraversalProfile
from metricflow_semantics.experimental.ordered_set import OrderedSet

logger = logging.getLogger(__name__)


@dataclass
class GraphTraversalResult(ABC):
    """A mixin dataclass that can be used to include the pathfinder counters in result objects."""

    traversal_profile: GraphTraversalProfile


@dataclass
class FindDescendantsResult(Generic[NodeT]):
    """A result object used to return the results of the `find_descendants` method."""

    # All nodes on any path from any source node to any target node.
    reachable_nodes: OrderedSet[NodeT]
    # The (sub)set of target nodes that were reachable from the source nodes.
    reachable_target_nodes: OrderedSet[NodeT]
    # The labels on edges / nodes that are on the path from any of the source nodes to any of the target nodes.
    labels_collected_during_traversal: OrderedSet[MetricFlowGraphLabel]
    # The number of iterations that were done to reach the target nodes or hit the iteration limit.
    finish_iteration_index: int
    # Maps the target node to the set of source nodes that could be reached.
    target_node_to_reachable_source_nodes: dict[NodeT, OrderedSet[NodeT]]


@dataclass
class FindAncestorsResult(Generic[NodeT]):
    """A result object used to return the results of the `find_ancestors` method."""

    # All nodes on any path from any source node to any target node.
    reachable_nodes: OrderedSet[NodeT]
    # The (sub)set of source nodes that were reachable from the source nodes.
    reachable_source_nodes: OrderedSet[NodeT]
    # Maps the source node to the target nodes that could be reached.
    source_node_to_reachable_target_nodes: Mapping[NodeT, OrderedSet[NodeT]]
    # The labels on edges / nodes that are on the path from any of the source nodes to any of the target nodes.
    labels_collected_during_traversal: OrderedSet[MetricFlowGraphLabel]
    # The number of iterations that were done to reach the target nodes or hit the iteration limit.
    finish_iteration_index: int



from __future__ import annotations

import logging
import threading
import time
from typing import ContextManager, Optional, Type, Union

from metricflow_semantics.collection_helpers.mf_type_aliases import ExceptionTracebackAnyType
from metricflow_semantics.helpers.time_helpers import PrettyDuration
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


class ExecutionTimer(ContextManager["ExecutionTimer"]):
    """A context manager for timing sections of code.

    This and associates classes are a WIP and may be removed.
    """

    def __init__(self, description: Optional[Union[str, LazyFormat]] = None) -> None:  # noqa: D107
        self._local_state = _ExecutionTimerLocalState(description)

    def __enter__(self) -> ExecutionTimer:  # noqa: D105
        description = self._local_state.description
        if description is not None:
            logger.info(LazyFormat(lambda: f"[  BEGIN  ] {description}"))
        self._local_state.start_time = time.perf_counter()
        return self

    def __exit__(  # noqa: D105
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[ExceptionTracebackAnyType],
    ) -> None:
        if self._local_state.start_time is None:
            logger.error(LazyFormat(lambda: f"{self.__class__.__name__} shouldn't exit context without entering."))
            return
        self._local_state.total_duration_for_completed_contexts += time.perf_counter() - self._local_state.start_time
        description = self._local_state.description
        if description is not None:
            total_duration = self._local_state.total_duration_for_completed_contexts
            logger.info(LazyFormat(lambda: f"[   END   ] {description} in {PrettyDuration(total_duration)}"))

    @property
    def total_duration(self) -> PrettyDuration:  # noqa: D102
        return PrettyDuration(self._local_state.total_duration_for_completed_contexts)


class _ExecutionTimerLocalState(threading.local):
    def __init__(self, description: Optional[Union[str, LazyFormat]]) -> None:  # noqa: D107
        self.start_time: Optional[float] = None
        self.total_duration_for_completed_contexts = 0.0
        self.description = description



from __future__ import annotations

import logging
from functools import cached_property
from typing import Optional

from dbt_semantic_interfaces.protocols import (
    MeasureAggregationParameters,
    NonAdditiveDimensionParameters,
    WhereFilterIntersection,
)
from dbt_semantic_interfaces.references import MetricReference
from dbt_semantic_interfaces.type_enums import AggregationType, TimeGranularity

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class SimpleMetricInputAggregation:
    """See `SimpleMetricInput`."""

    percentile: Optional[float] = None
    use_discrete_percentile: bool = False
    use_approximate_percentile: bool = False

    @staticmethod
    def create_from_pydantic(  # noqa: D102
        aggregation: Optional[MeasureAggregationParameters],
    ) -> Optional[SimpleMetricInputAggregation]:
        if aggregation is None:
            return None
        return SimpleMetricInputAggregation(
            percentile=aggregation.percentile,
            use_discrete_percentile=aggregation.use_discrete_percentile,
            use_approximate_percentile=aggregation.use_approximate_percentile,
        )


@fast_frozen_dataclass()
class SimpleMetricInputNonAdditiveDimension:
    """See `SimpleMetricInput`."""

    name: str
    window_choice: AggregationType
    window_groupings: AnyLengthTuple[str]

    @staticmethod
    def create_from_non_additive_dimension(  # noqa: D102
        non_additive_dimension: Optional[NonAdditiveDimensionParameters],
    ) -> Optional[SimpleMetricInputNonAdditiveDimension]:
        if non_additive_dimension is None:
            return None

        return SimpleMetricInputNonAdditiveDimension(
            name=non_additive_dimension.name,
            window_choice=non_additive_dimension.window_choice,
            window_groupings=tuple(non_additive_dimension.window_groupings),
        )


@fast_frozen_dataclass()
class SimpleMetricInput:
    """Represents the input arguments to construct a simple metric.

    This class should contain all relevant values from the semantic manifest for a simple metric. i.e. use this
    instead of fetching the `Metric` from the manifest.

    These fields are based on the current Pydantic classes, but they can be better organized / consolidated.
    """

    name: str
    agg: AggregationType
    expr: str
    agg_params: Optional[SimpleMetricInputAggregation]
    non_additive_dimension: Optional[SimpleMetricInputNonAdditiveDimension]
    agg_time_dimension_name: str
    agg_time_dimension_grain: TimeGranularity
    model_id: SemanticModelId
    join_to_timespine: bool
    fill_nulls_with: Optional[int]
    # This is the filter in the definition of the simple metric.
    filter: WhereFilterIntersection

    @cached_property
    def metric_reference(self) -> MetricReference:  # noqa: D102
        return MetricReference(self.name)



from __future__ import annotations

import logging
import typing
from abc import ABC, abstractmethod

if typing.TYPE_CHECKING:
    from metricflow_semantics.experimental.mf_graph.mf_graph import MetricFlowGraph

logger = logging.getLogger(__name__)


class MetricFlowGraphFormatter(ABC):
    """Interface for a graph-to-text formatter."""

    @abstractmethod
    def format_graph(self, graph: MetricFlowGraph) -> str:
        """Format the given graph to text."""
        raise NotImplementedError



from __future__ import annotations

import logging
from typing import Optional

from dbt_semantic_interfaces.type_enums import DatePart

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe import IndexedDunderName
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class EntityKeyQueryForGroupByMetric:
    """Represents a possible entity-key query for group-by metric.

    e.g for the `bookings_source` semantic model, `booking__listing` is a possible entity-key query.
    """

    entity_key_query: IndexedDunderName
    derived_from_model_ids: AnyLengthTuple[SemanticModelId]


@fast_frozen_dataclass()
class DunderNameDescriptor:
    """A descriptor used in `DunderNameTrie` that can be used to build a `AnnotatedSpec` later.

    The fields here are analogous to the ones in `AnnotatedSpec`.

    These descriptors are used in trie-based data structures that describe the available group-by items for a metric (
    see `DunderNameTrie`). In these structures, the dunder name maps to this descriptor.

    Some consolidation is needed between similar descriptor types.
    """

    element_type: LinkableElementType
    time_grain: Optional[ExpandedTimeGranularity]
    date_part: Optional[DatePart]

    element_properties: AnyLengthTuple[GroupByItemProperty]
    origin_model_ids: AnyLengthTuple[SemanticModelId]
    derived_from_model_ids: AnyLengthTuple[SemanticModelId]
    entity_key_queries_for_group_by_metric: AnyLengthTuple[EntityKeyQueryForGroupByMetric]

    def merge(self, other: DunderNameDescriptor) -> DunderNameDescriptor:
        """Combine the properties of this with the other descriptor.

        To generate the trie for a metric from the corresponding trie for the constituent measures, there
        needs to be a mechanism for combining them to reflect properties like `derived_from_model_ids`.

        For example, a metric like `bookings_per_view` would be based on the `booking` measure (in the
        `booking_source` model) and the `views` measure (in the `views_source` model). Combining the trie from each
        model should result in descriptors that include both `views_source` and `bookings_source`.
        """
        # Fields like `element_type` and `time_grain` should be the same for both descriptors if there were no bugs
        # with the construction of the trie. Considering adding a check.
        return DunderNameDescriptor(
            element_type=self.element_type,
            time_grain=self.time_grain,
            date_part=self.date_part,
            element_properties=self.element_properties + other.element_properties,
            origin_model_ids=self.origin_model_ids + other.origin_model_ids,
            derived_from_model_ids=self.derived_from_model_ids + other.derived_from_model_ids,
            entity_key_queries_for_group_by_metric=self.entity_key_queries_for_group_by_metric,
        )

    def merge_derived_from_model_ids(
        self, derived_from_model_ids: AnyLengthTuple[SemanticModelId]
    ) -> DunderNameDescriptor:
        """Return a new descriptor that is a copy of this but with additional `derived_from_model_ids`."""
        return DunderNameDescriptor(
            element_type=self.element_type,
            time_grain=self.time_grain,
            date_part=self.date_part,
            element_properties=self.element_properties,
            origin_model_ids=self.origin_model_ids,
            derived_from_model_ids=self.derived_from_model_ids + derived_from_model_ids,
            entity_key_queries_for_group_by_metric=self.entity_key_queries_for_group_by_metric,
        )

    def is_mergeable(self, other: DunderNameDescriptor) -> bool:
        """Returns True if this descriptor can form a union with another descriptor."""
        return (
            self.element_type is other.element_type
            and self.time_grain == other.time_grain
            and self.date_part is other.date_part
            and self.entity_key_queries_for_group_by_metric == other.entity_key_queries_for_group_by_metric
        )



from __future__ import annotations

import logging
from functools import cached_property
from typing import Optional, Set

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums import DatePart, TimeGranularity

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.metricflow_exception import MetricFlowInternalError
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet, OrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


IndexedDunderName = AnyLengthTuple[str]


@fast_frozen_dataclass()
class AttributeRecipe:
    """The recipe for computing an attribute by following a path in the semantic graph."""

    indexed_dunder_name: IndexedDunderName = ()
    joined_model_ids: AnyLengthTuple[SemanticModelId] = ()
    element_properties: FrozenOrderedSet[GroupByItemProperty] = FrozenOrderedSet()
    entity_link_names: AnyLengthTuple[str] = ()

    element_type: Optional[LinkableElementType] = None
    # Maps to the time grain set for a time dimension in a semantic model
    source_time_grain: Optional[TimeGranularity] = None
    # Maps to the attribute's time grain or date part.
    recipe_time_grain: Optional[ExpandedTimeGranularity] = None
    recipe_date_part: Optional[DatePart] = None

    @cached_property
    def dunder_name_elements_set(self) -> Set[str]:
        """The elements of a dunder name as a set for fast repeated-element checks."""
        return set(self.indexed_dunder_name)

    @cached_property
    def joined_model_id_set(self) -> Set[SemanticModelId]:
        """The joined semantic model IDs as a set for fast repeated-model checks."""
        return set(self.joined_model_ids)

    @staticmethod
    def create(initial_step: AttributeRecipeStep) -> AttributeRecipe:  # noqa: D102
        dunder_name_elements: AnyLengthTuple[str] = ()
        if initial_step.add_dunder_name_element is not None:
            dunder_name_elements = (initial_step.add_dunder_name_element,)
        entity_link_names: AnyLengthTuple[str] = ()
        if initial_step.add_entity_link is not None:
            entity_link_names = (initial_step.add_entity_link,)

        models_in_join: AnyLengthTuple[SemanticModelId] = ()

        add_model_join = initial_step.add_model_join
        if add_model_join is not None:
            models_in_join = models_in_join + (add_model_join,)

        return AttributeRecipe(
            indexed_dunder_name=dunder_name_elements,
            joined_model_ids=models_in_join,
            element_properties=FrozenOrderedSet(initial_step.add_properties or ()),
            element_type=initial_step.set_element_type,
            entity_link_names=entity_link_names,
            source_time_grain=initial_step.set_source_time_grain,
            recipe_time_grain=initial_step.set_time_grain_access,
            recipe_date_part=initial_step.set_date_part_access,
        )

    @cached_property
    def last_model_id(self) -> Optional[SemanticModelId]:
        """The last model ID that was added to the join."""
        if self.joined_model_ids:
            return None

        return tuple(self.joined_model_ids)[-1]

    def append_step(self, recipe_step: AttributeRecipeStep) -> AttributeRecipe:
        """Add a step to the end of the recipe."""
        dundered_name_elements = self.indexed_dunder_name
        if recipe_step.add_dunder_name_element is not None:
            dundered_name_elements = dundered_name_elements + (recipe_step.add_dunder_name_element,)
        entity_link_names = self.entity_link_names
        if recipe_step.add_entity_link is not None:
            entity_link_names = entity_link_names + (recipe_step.add_entity_link,)

        models_in_join = self.joined_model_ids
        join_model = recipe_step.add_model_join

        if join_model is not None:
            models_in_join = models_in_join + (join_model,)

        return AttributeRecipe(
            indexed_dunder_name=dundered_name_elements,
            joined_model_ids=models_in_join,
            element_properties=self.element_properties.union(recipe_step.add_properties)
            if recipe_step.add_properties is not None
            else self.element_properties,
            element_type=recipe_step.set_element_type or self.element_type,
            entity_link_names=entity_link_names,
            source_time_grain=recipe_step.set_source_time_grain or self.source_time_grain,
            recipe_time_grain=recipe_step.set_time_grain_access or self.recipe_time_grain,
            recipe_date_part=recipe_step.set_date_part_access or self.recipe_date_part,
        )

    def push_step(self, recipe_step: AttributeRecipeStep) -> AttributeRecipe:
        """Add a step to the beginning of the recipe."""
        dundered_name_elements = self.indexed_dunder_name
        if recipe_step.add_dunder_name_element is not None:
            dundered_name_elements = (recipe_step.add_dunder_name_element,) + dundered_name_elements
        entity_link_names = self.entity_link_names
        if recipe_step.add_entity_link is not None:
            entity_link_names = (recipe_step.add_entity_link,) + entity_link_names
        models_in_join = self.joined_model_ids
        add_model_join = recipe_step.add_model_join
        if add_model_join is not None:
            if len(models_in_join) == 0:
                models_in_join = (add_model_join,)
            else:
                models_in_join = (add_model_join,) + models_in_join

        return AttributeRecipe(
            indexed_dunder_name=dundered_name_elements,
            joined_model_ids=models_in_join,
            element_properties=FrozenOrderedSet(recipe_step.add_properties).union(self.element_properties)
            if recipe_step.add_properties is not None
            else self.element_properties,
            element_type=self.element_type or recipe_step.set_element_type,
            entity_link_names=entity_link_names,
            source_time_grain=self.source_time_grain or recipe_step.set_source_time_grain,
            recipe_time_grain=self.recipe_time_grain or recipe_step.set_time_grain_access,
            recipe_date_part=self.recipe_date_part or recipe_step.set_date_part_access,
        )

    def push_steps(self, *updates: AttributeRecipeStep) -> AttributeRecipe:
        """See `push_step`."""
        result = self
        for update in updates:
            result = result.push_step(update)
        return result

    def resolve_complete_properties(self) -> OrderedSet[GroupByItemProperty]:
        """Resolve the complete set of `GroupByItemProperty` for this recipe.

        While many properties were set by recipe steps during traversal, some need to be resolved at the end as it
        is easier / faster to determine at the end.
        """
        element_type = self.element_type

        if element_type is None:
            raise ValueError(LazyFormat("Recipe is missing the element type", recipe=self))

        properties = MutableOrderedSet(self.element_properties)

        model_ids = self.joined_model_ids
        model_id_count = len(model_ids)
        if model_id_count == 0:
            if GroupByItemProperty.METRIC_TIME not in properties:
                raise ValueError(LazyFormat("Recipe is missing context on accessed semantic models", recipe=self))
        elif model_id_count == 1:
            if element_type is not LinkableElementType.METRIC and GroupByItemProperty.METRIC_TIME not in properties:
                properties.add(GroupByItemProperty.LOCAL)
        elif model_id_count == 2:
            properties.add(GroupByItemProperty.JOINED)
        elif model_id_count >= 3:
            properties.update(
                (
                    GroupByItemProperty.JOINED,
                    GroupByItemProperty.MULTI_HOP,
                )
            )
        else:
            raise MetricFlowInternalError(
                LazyFormat("Reached unhandled case", model_id_count=model_id_count, recipe=self)
            )

        # Add `DERIVED_TIME_GRANULARITY` if the grain is different from the element's grain.
        source_time_grain = self.source_time_grain
        recipe_time_grain = self.recipe_time_grain
        if source_time_grain is not None:
            if recipe_time_grain is None and self.recipe_date_part is None:
                raise ValueError(
                    LazyFormat(
                        "Recipe has a source time-grain, but no recipe time-grain or recipe date-part", recipe=self
                    )
                )
            if recipe_time_grain is not None and source_time_grain is not recipe_time_grain.base_granularity:
                properties.add(GroupByItemProperty.DERIVED_TIME_GRANULARITY)

        return properties

    def resolve_element_name(self) -> Optional[str]:
        """Resolve the element name.

        Currently, the recipe stores the dunder-name elements (e.g. ["metric_time", "day"]), but not the element name.
        Since the position of the element name in the list depends on the type of element, this method helps to resolve
        that.
        """
        element_type = self.element_type

        # Incomplete recipe.
        if element_type is None:
            return None

        dunder_name_elements = self.indexed_dunder_name
        dunder_name_element_count = len(dunder_name_elements)

        # Incomplete recipe.
        if dunder_name_element_count == 0:
            return None

        if element_type is LinkableElementType.TIME_DIMENSION:
            # e.g. ['metric_time']
            if dunder_name_element_count == 1:
                return dunder_name_elements[-1]
            # e.g. ['metric_time', 'day']
            else:
                return dunder_name_elements[-2]
        elif (
            element_type is LinkableElementType.ENTITY
            or element_type is LinkableElementType.DIMENSION
            or element_type is LinkableElementType.TIME_DIMENSION
            or element_type is LinkableElementType.METRIC
        ):
            return dunder_name_elements[-1]
        else:
            assert_values_exhausted(element_type)



from __future__ import annotations

import logging

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.singleton import Singleton

logger = logging.getLogger(__name__)


@fast_frozen_dataclass(order=False)
class CumulativeMetricLabel(MetricFlowGraphLabel, Singleton):
    """Label for an edge from a cumulative-metric node to a simple-metric node.

    This label is helpful for addressing special cases with cumulative metrics (e.g. ability to query by date part).
    """

    @classmethod
    def get_instance(cls) -> CumulativeMetricLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class DenyDatePartLabel(MetricFlowGraphLabel, Singleton):
    """Label for edges that, when added to a path, should not allow querying of the date part."""

    @classmethod
    def get_instance(cls) -> DenyDatePartLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class DenyEntityKeyQueryResolutionLabel(MetricFlowGraphLabel, Singleton):
    """Label for edges that, when added to a path, should not allow querying of only the entity-key attributes.

    e.g. for time-offset metrics, the successor edges have this label as those metrics must be queried with
    `metric_time` and it is not possible to query the metric only for entity-key attributes.
    """

    @classmethod
    def get_instance(cls) -> DenyEntityKeyQueryResolutionLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class ConversionMetricLabel(MetricFlowGraphLabel, Singleton):
    """Label for successor edges from a conversion metric to the input conversion metric node."""

    @classmethod
    def get_instance(cls) -> ConversionMetricLabel:  # noqa: D102
        return cls._get_instance()


@fast_frozen_dataclass(order=False)
class DenyVisibleAttributesLabel(MetricFlowGraphLabel, Singleton):
    """Label for edges that, when added to a path, should not affect visibility of group-by items.

    e.g. for conversion metrics, the edge from the conversion-metric node to the conversion-measure node is given this
    label as the conversion measure does not affect the available group-by items for the metric.
    """

    @classmethod
    def get_instance(cls) -> DenyVisibleAttributesLabel:  # noqa: D102
        return cls._get_instance()



from __future__ import annotations

import logging
from functools import cached_property
from typing import Optional

from typing_extensions import override

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.comparable import ComparisonKey
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, OrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import SemanticGraphEdge, SemanticGraphNode
from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext

logger = logging.getLogger(__name__)


@fast_frozen_dataclass(order=False)
class JoinToModelEdge(SemanticGraphEdge):
    """Edge that describes joining a model on the right side."""

    right_model_id: SemanticModelId

    @staticmethod
    def create(  # noqa: D102
        tail_node: SemanticGraphNode,
        head_node: SemanticGraphNode,
        right_model_id: SemanticModelId,
    ) -> JoinToModelEdge:
        return JoinToModelEdge(
            tail_node=tail_node,
            head_node=head_node,
            right_model_id=right_model_id,
        )

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return (self.tail_node, self.head_node, self.right_model_id)

    @cached_property
    @override
    def inverse(self) -> JoinToModelEdge:
        return JoinToModelEdge.create(
            tail_node=self.tail_node,
            head_node=self.head_node,
            right_model_id=self.right_model_id,
        )

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return AttributeRecipeStep(add_model_join=self.right_model_id)


_EMPTY_LABEL_SET: FrozenOrderedSet[MetricFlowGraphLabel] = FrozenOrderedSet()
_EMPTY_RECIPE_STEP = AttributeRecipeStep()


@fast_frozen_dataclass(order=False)
class MetricDefinitionEdge(SemanticGraphEdge):
    """An edge that points from a metric to the inputs for the metric."""

    additional_labels: FrozenOrderedSet[MetricFlowGraphLabel]
    _recipe_step: AttributeRecipeStep

    @staticmethod
    def create(  # noqa: D102
        tail_node: SemanticGraphNode,
        head_node: SemanticGraphNode,
        additional_labels: Optional[FrozenOrderedSet[MetricFlowGraphLabel]] = None,
        recipe_step: Optional[AttributeRecipeStep] = None,
    ) -> MetricDefinitionEdge:
        return MetricDefinitionEdge(
            tail_node=tail_node,
            head_node=head_node,
            additional_labels=additional_labels if additional_labels is not None else _EMPTY_LABEL_SET,
            _recipe_step=recipe_step if recipe_step is not None else _EMPTY_RECIPE_STEP,
        )

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return (
            self.tail_node,
            self.head_node,
        )

    @cached_property
    @override
    def inverse(self) -> MetricDefinitionEdge:
        return MetricDefinitionEdge.create(
            tail_node=self.head_node,
            head_node=self.tail_node,
        )

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return self._recipe_step

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        formatter = format_context.formatter
        return formatter.pretty_format_object_by_parts(
            class_name=self.__class__.__name__,
            field_mapping={
                "tail_node": self.tail_node,
                "head_node": self.head_node,
                "labels": self.labels,
                "recipe_step": self.recipe_step_to_append,
            },
        )

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return FrozenOrderedSet(self.additional_labels)


@fast_frozen_dataclass(order=False)
class EntityRelationshipEdge(SemanticGraphEdge):
    """Represents a relationship between entity nodes."""

    _recipe_update: AttributeRecipeStep

    @staticmethod
    def create(  # noqa: D102
        tail_node: SemanticGraphNode,
        head_node: SemanticGraphNode,
        recipe_update: AttributeRecipeStep = _EMPTY_RECIPE_STEP,
    ) -> EntityRelationshipEdge:
        return EntityRelationshipEdge(
            tail_node=tail_node,
            head_node=head_node,
            _recipe_update=recipe_update,
        )

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return (self.tail_node, self.head_node, self._recipe_update)

    @cached_property
    @override
    def inverse(self) -> EntityRelationshipEdge:
        return EntityRelationshipEdge.create(
            tail_node=self.head_node,
            head_node=self.tail_node,
            recipe_update=self._recipe_update,
        )

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return self._recipe_update


@fast_frozen_dataclass(order=False)
class EntityAttributeEdge(SemanticGraphEdge):
    """Edge from entity nodes to attribute nodes."""

    _recipe_step: AttributeRecipeStep

    @staticmethod
    def create(  # noqa: D102
        tail_node: SemanticGraphNode,
        head_node: SemanticGraphNode,
        recipe_step: Optional[AttributeRecipeStep] = None,
    ) -> EntityAttributeEdge:
        return EntityAttributeEdge(
            tail_node=tail_node,
            head_node=head_node,
            _recipe_step=recipe_step or _EMPTY_RECIPE_STEP,
        )

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return (self.tail_node, self.head_node, self._recipe_step)

    @cached_property
    @override
    def inverse(self) -> EntityAttributeEdge:
        return EntityAttributeEdge.create(
            tail_node=self.head_node,
            head_node=self.tail_node,
            recipe_step=self._recipe_step,
        )

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return self._recipe_step



from __future__ import annotations

import logging
from dataclasses import dataclass
from functools import cached_property
from typing import Iterable

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from typing_extensions import override

from metricflow.sql.sql_plan import SqlSelectColumn

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SelectColumnSet(Mergeable):
    """A set of SQL select columns that represent the different instance types in a data set.

    TODO: Evaluate using a single field instead of one for every instance type.
    """

    metric_columns: AnyLengthTuple[SqlSelectColumn]
    simple_metric_input_columns: AnyLengthTuple[SqlSelectColumn]
    dimension_columns: AnyLengthTuple[SqlSelectColumn]
    time_dimension_columns: AnyLengthTuple[SqlSelectColumn]
    entity_columns: AnyLengthTuple[SqlSelectColumn]
    group_by_metric_columns: AnyLengthTuple[SqlSelectColumn]
    metadata_columns: AnyLengthTuple[SqlSelectColumn]

    @staticmethod
    def create(  # noqa: D102
        metric_columns: Iterable[SqlSelectColumn] = (),
        simple_metric_input_columns: Iterable[SqlSelectColumn] = (),
        dimension_columns: Iterable[SqlSelectColumn] = (),
        time_dimension_columns: Iterable[SqlSelectColumn] = (),
        entity_columns: Iterable[SqlSelectColumn] = (),
        group_by_metric_columns: Iterable[SqlSelectColumn] = (),
        metadata_columns: Iterable[SqlSelectColumn] = (),
    ) -> SelectColumnSet:
        metric_columns = tuple(metric_columns)
        simple_metric_input_columns = tuple(simple_metric_input_columns)
        dimension_columns = tuple(dimension_columns)
        time_dimension_columns = tuple(time_dimension_columns)
        entity_columns = tuple(entity_columns)
        group_by_metric_columns = tuple(group_by_metric_columns)
        metadata_columns = tuple(metadata_columns)

        return SelectColumnSet(
            metric_columns=tuple(metric_columns),
            simple_metric_input_columns=tuple(simple_metric_input_columns),
            dimension_columns=tuple(dimension_columns),
            time_dimension_columns=tuple(time_dimension_columns),
            entity_columns=tuple(entity_columns),
            group_by_metric_columns=tuple(group_by_metric_columns),
            metadata_columns=tuple(metadata_columns),
        )

    @cached_property
    def columns_in_default_order(self) -> AnyLengthTuple[SqlSelectColumn]:  # noqa: D102
        return (
            # This order was chosen to match the column sequence data consumers typically prefer.
            self.time_dimension_columns
            + self.entity_columns
            + self.dimension_columns
            + self.group_by_metric_columns
            + self.metric_columns
            + self.simple_metric_input_columns
            + self.metadata_columns
        )

    @override
    def merge(self, other_set: SelectColumnSet) -> SelectColumnSet:
        """Combine the select columns by type."""
        return SelectColumnSet.create(
            metric_columns=self.metric_columns + other_set.metric_columns,
            simple_metric_input_columns=self.simple_metric_input_columns + other_set.simple_metric_input_columns,
            dimension_columns=self.dimension_columns + other_set.dimension_columns,
            time_dimension_columns=self.time_dimension_columns + other_set.time_dimension_columns,
            entity_columns=self.entity_columns + other_set.entity_columns,
            group_by_metric_columns=self.group_by_metric_columns + other_set.group_by_metric_columns,
            metadata_columns=self.metadata_columns + other_set.metadata_columns,
        )

    @classmethod
    @override
    def empty_instance(cls) -> SelectColumnSet:
        return SelectColumnSet.create()



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Hashable, MutableSet, Set
from functools import cached_property
from typing import Generic, Iterable, Iterator, Optional, TypeVar

from typing_extensions import Self, override

from metricflow_semantics.collection_helpers.mf_type_aliases import HashableT, HashableT_co
from metricflow_semantics.experimental.comparison_helpers import ComparisonOtherType

logger = logging.getLogger(__name__)

OrderedSetT = TypeVar("OrderedSetT", bound="OrderedSet", covariant=True)


class OrderedSet(Generic[HashableT_co], Set[HashableT_co], ABC):
    """Set where the iteration order is the insertion order.

    * Having a consistent iteration order is helpful for ensuring consistency in tests and snapshot generation without
      sorting.
    * Since this is meant to be used for consistency in iteration, eq / hash behavior does not depend on order.
    * Considering using the `ordered-set` package instead, but it would add a dependency and this implementation is
      short.
    """

    def __init__(
        self,
        iterable: Optional[Iterable[HashableT_co]] = None,
        _set_as_dict: Optional[dict[HashableT_co, None]] = None,
    ) -> None:
        """Initializer.

        Args:
            iterable: Create the set using the given iterable.
            _set_as_dict: For internal use cases: Use this exact object for the internal dict. Useful in copy-like
            operations as `dict.copy()` is faster than initializing a new dictionary with items. The performance
            difference is only relevant in tight loops.
        """
        self._set_as_dict = _set_as_dict if _set_as_dict is not None else {}
        if iterable is not None:
            self._set_as_dict.update({item: None for item in iterable})

    def difference(self, *others: Iterable[HashableT_co]) -> Self:  # noqa: D102
        items_to_remove = set(item for other in others for item in other)
        return self.__class__(_set_as_dict={item: None for item in self._set_as_dict if item not in items_to_remove})

    def intersection(self, *others: Iterable[HashableT_co]) -> Self:  # noqa: D102
        common_keys = set(self._set_as_dict).intersection(*others)
        return self.__class__(
            _set_as_dict={
                key: value
                # Iterating through the dict instead of `common_keys` to retain order.
                for key, value in self._set_as_dict.items()
                if key in common_keys
            }
        )

    def union(self, *others: Iterable[HashableT_co]) -> Self:  # noqa: D102
        return self.__class__(_set_as_dict={item: None for other_set in (self,) + tuple(others) for item in other_set})

    @override
    def __contains__(self, obj: ComparisonOtherType) -> bool:
        return obj in self._set_as_dict

    @override
    def __len__(self) -> int:
        return len(self._set_as_dict)

    @override
    def __iter__(self) -> Iterator[HashableT_co]:
        return iter(self._set_as_dict)

    @override
    def __str__(self) -> str:
        return "{" + ", ".join(str(item) for item in self) + "}"

    @abstractmethod
    def as_mutable(self) -> MutableOrderedSet[HashableT_co]:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def as_frozen(self) -> FrozenOrderedSet[HashableT_co]:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def copy(self) -> Self:
        """Return a shallow copy of this set."""
        raise NotImplementedError


class FrozenOrderedSet(Generic[HashableT_co], OrderedSet[HashableT_co], Hashable):
    """A frozen version of `OrderedSet` that can be hashed."""

    def __init__(  # noqa: D107
        self,
        iterable: Optional[Iterable[HashableT_co]] = None,
        _set_as_dict: Optional[dict[HashableT_co, None]] = None,
    ) -> None:
        super().__init__(iterable, _set_as_dict)

    @cached_property
    def _cached_hash_value(self) -> int:
        return hash(tuple(self._set_as_dict))

    @override
    def __hash__(self) -> int:
        return self._cached_hash_value

    @override
    def as_mutable(self) -> MutableOrderedSet[HashableT_co]:
        return MutableOrderedSet(_set_as_dict=self._set_as_dict.copy())

    @override
    def as_frozen(self) -> FrozenOrderedSet[HashableT_co]:
        return self

    @override
    def copy(self) -> FrozenOrderedSet[HashableT_co]:
        return FrozenOrderedSet(_set_as_dict=self._set_as_dict)


class MutableOrderedSet(Generic[HashableT], OrderedSet[HashableT], MutableSet[HashableT]):
    """An ordered set that can be modified."""

    def update(self, *iterables: Iterable[HashableT]) -> None:  # noqa: D102
        for iterable in iterables:
            for item in iterable:
                self._set_as_dict[item] = None

    def intersection_update(self, *others: Iterable[HashableT]) -> None:  # noqa: D102
        intersected_items: set[HashableT] = set(self._set_as_dict).intersection(*others)
        self._set_as_dict = {item: None for item in self._set_as_dict if item in intersected_items}

    def difference_update(self, *others: Iterable[HashableT]) -> None:  # noqa: D102
        items_to_remove = {item for other in others for item in other}
        self._set_as_dict = {item: None for item in self._set_as_dict if item not in items_to_remove}

    @override
    def add(self, value: HashableT) -> None:
        self._set_as_dict[value] = None

    @override
    def discard(self, value: HashableT) -> None:
        self._set_as_dict.pop(value, None)

    @override
    def as_mutable(self) -> MutableOrderedSet[HashableT]:
        return self

    @override
    def as_frozen(self) -> FrozenOrderedSet[HashableT]:
        return FrozenOrderedSet(_set_as_dict=self._set_as_dict.copy())

    def clear(self) -> None:
        """Remove all items from this set."""
        self._set_as_dict.clear()

    def pop(self) -> HashableT:
        """Removes and returns the first item in the set. Raises `KeyError` if empty (same as `set()`)."""
        try:
            item = next(iter(self._set_as_dict))
            del self._set_as_dict[item]
            return item
        except StopIteration:
            raise KeyError("Can't pop an item as the set is empty.")

    @override
    def copy(self) -> MutableOrderedSet[HashableT]:
        return MutableOrderedSet(_set_as_dict=self._set_as_dict.copy())



from __future__ import annotations

import logging
from functools import cached_property
from typing import Final, Iterable, Optional

from typing_extensions import override

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.mf_logging.pretty_formatter import PrettyFormatContext

logger = logging.getLogger(__name__)


class PrettyDuration(MetricFlowPrettyFormattable):
    """Wrapper to format durations using the MF pretty-printer."""

    def __init__(self, seconds: float, decimals: int = 2) -> None:  # noqa: D107
        self.seconds: Final[float] = seconds
        if decimals < 0:
            logger.error(LazyFormat("`decimals` argument must not be negative. Using 0 instead.", decimals=decimals))
            decimals = 0
        self._decimals = decimals

    @cached_property
    def _pretty_seconds(self) -> str:
        return f"{self.seconds:.{self._decimals}f} s"

    @override
    def __str__(self) -> str:
        return self._pretty_seconds

    @override
    def pretty_format(self, format_context: PrettyFormatContext) -> Optional[str]:
        return self._pretty_seconds

    @staticmethod
    def sum(durations: Iterable[PrettyDuration]) -> PrettyDuration:  # noqa: D102
        return PrettyDuration(
            sum(duration.seconds for duration in durations),
            max(duration._decimals for duration in durations),
        )



from __future__ import annotations

import logging
from collections.abc import Mapping, Set
from typing import Sequence

from metricflow_semantics.collection_helpers.mf_type_aliases import KeyT

logger = logging.getLogger(__name__)


def mf_common_keys(dicts: Sequence[Mapping[KeyT, object]]) -> Set[KeyT]:
    """Returns the set of keys that are common to all dictionaries."""
    if not dicts:
        return set()

    first_dict = dicts[0]
    return set(first_dict).intersection(*[other_dict for other_dict in dicts[1:]])



from __future__ import annotations

import logging
from functools import cached_property
from typing import Dict, List, Mapping, Optional, Sequence, Set, Tuple

from dbt_semantic_interfaces.protocols.entity import Entity
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    DimensionReference,
    EntityReference,
    SemanticModelElementReference,
    SemanticModelReference,
    TimeDimensionReference,
)
from dbt_semantic_interfaces.type_enums import DimensionType

from metricflow_semantics.errors.error_classes import InvalidSemanticModelError
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.dimension_lookup import DimensionLookup
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.time_dimension_spec import DEFAULT_TIME_GRANULARITY, TimeDimensionSpec
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


class SemanticModelLookup:
    """Tracks semantic information for semantic models held in a set of SemanticModelContainers."""

    def __init__(
        self, semantic_manifest: SemanticManifest, custom_granularities: Dict[str, ExpandedTimeGranularity]
    ) -> None:
        """Initializer.

        Args:
            semantic_manifest: the semantic manifest used for loading semantic model definitions
        """
        self.custom_granularities = custom_granularities
        self._dimension_index: Dict[DimensionReference, List[SemanticModel]] = {}
        self.entity_index: Dict[EntityReference, List[SemanticModel]] = {}

        self._dimension_ref_to_spec: Dict[DimensionReference, DimensionSpec] = {}
        self._entity_ref_to_spec: Dict[EntityReference, EntitySpec] = {}

        self._semantic_model_reference_to_semantic_model: Dict[SemanticModelReference, SemanticModel] = {}
        sorted_semantic_models = sorted(
            semantic_manifest.semantic_models, key=lambda semantic_model: semantic_model.name
        )
        for semantic_model in sorted_semantic_models:
            self._add_semantic_model(semantic_model)

        self._dimension_lookup = DimensionLookup(sorted_semantic_models)

    @cached_property
    def custom_granularity_names(self) -> Tuple[str, ...]:
        """Returns all the custom_granularity names."""
        return tuple(self.custom_granularities.keys())

    def get_dimension_references(self) -> Sequence[DimensionReference]:
        """Retrieve all dimension references from the collection of semantic models."""
        return tuple(self._dimension_index.keys())

    def get_entity_references(self) -> Sequence[EntityReference]:
        """Retrieve all entity references from the collection of semantic models."""
        return list(self.entity_index.keys())

    def get_entity_in_semantic_model(self, ref: SemanticModelElementReference) -> Optional[Entity]:
        """Retrieve the entity matching the element -> semantic model mapping, if any."""
        semantic_model = self.get_by_reference(ref.semantic_model_reference)
        if not semantic_model:
            return None

        for entity in semantic_model.entities:
            if entity.reference.element_name == ref.element_name:
                return entity

        return None

    def get_by_reference(self, semantic_model_reference: SemanticModelReference) -> Optional[SemanticModel]:
        """Retrieve the semantic model object matching the input semantic model reference, if any."""
        return self._semantic_model_reference_to_semantic_model.get(semantic_model_reference)

    def _add_semantic_model(self, semantic_model: SemanticModel) -> None:
        """Add semantic model semantic information, validating consistency with existing semantic models."""
        errors = []

        if semantic_model.reference in self._semantic_model_reference_to_semantic_model:
            errors.append(f"Semantic model {semantic_model.reference} already added.")

        if len(errors) > 0:
            raise InvalidSemanticModelError(
                LazyFormat(
                    "Got errors adding the given semantic model.", semantic_model=semantic_model.name, errors=errors
                )
            )

        for dim in semantic_model.dimensions:
            semantic_models_for_dimension = self._dimension_index.get(dim.reference, []) + [semantic_model]
            self._dimension_index[dim.reference] = semantic_models_for_dimension

            if not StructuredLinkableSpecName.from_name(
                qualified_name=dim.name, custom_granularity_names=self.custom_granularity_names
            ).is_element_name:
                # TODO: [custom granularity] change this to an assertion once we're sure there aren't exceptions
                logger.warning(
                    LazyFormat(
                        lambda: f"Dimension name `{dim.name}` contains annotations, but this name should be the plain element name "
                        "from the original model. This should have been blocked by validation!"
                    )
                )

            if dim.type is DimensionType.TIME:
                defined_granularity = dim.type_params.time_granularity if dim.type_params else DEFAULT_TIME_GRANULARITY
                assert dim.time_dimension_reference, f"Time dimension {dim} does not have a time dimension reference"
                self._dimension_ref_to_spec[dim.time_dimension_reference] = TimeDimensionSpec(
                    element_name=dim.name,
                    entity_links=(),
                    time_granularity=ExpandedTimeGranularity.from_time_granularity(defined_granularity),
                )
            else:
                self._dimension_ref_to_spec[dim.reference] = DimensionSpec(element_name=dim.name, entity_links=())

        for entity in semantic_model.entities:
            semantic_models_for_entity = self.entity_index.get(entity.reference, []) + [semantic_model]
            self.entity_index[entity.reference] = semantic_models_for_entity

            self._entity_ref_to_spec[entity.reference] = EntitySpec(element_name=entity.name, entity_links=())

        self._semantic_model_reference_to_semantic_model[semantic_model.reference] = semantic_model

    def get_semantic_models_for_entity(self, entity_reference: EntityReference) -> Set[SemanticModel]:
        """Return all semantic models associated with an entity reference."""
        return set(self.entity_index.get(entity_reference, []))

    def get_semantic_models_for_dimension(self, dimension_reference: DimensionReference) -> Set[SemanticModel]:
        """Return all semantic models associated with a dimension reference."""
        return set(self._dimension_index.get(dimension_reference, []))

    def get_element_spec_for_name(self, element_name: str) -> LinkableInstanceSpec:
        """Returns the spec for the given name of a linkable element (dimension or entity)."""
        if TimeDimensionReference(element_name=element_name) in self._dimension_ref_to_spec:
            return self._dimension_ref_to_spec[TimeDimensionReference(element_name=element_name)]
        elif DimensionReference(element_name=element_name) in self._dimension_ref_to_spec:
            return self._dimension_ref_to_spec[DimensionReference(element_name=element_name)]
        elif EntityReference(element_name=element_name) in self._entity_ref_to_spec:
            return self._entity_ref_to_spec[EntityReference(element_name=element_name)]
        else:
            raise ValueError(f"Unable to find linkable element {element_name} in manifest")

    @property
    def dimension_lookup(self) -> DimensionLookup:  # noqa: D102
        return self._dimension_lookup

    @cached_property
    def model_reference_to_model(self) -> Mapping[SemanticModelReference, SemanticModel]:  # noqa: D102
        return {
            model_reference: self._semantic_model_reference_to_semantic_model[model_reference]
            for model_reference in sorted(self._semantic_model_reference_to_semantic_model)
        }



from __future__ import annotations

import logging
from collections.abc import Sequence
from typing import Type

from typing_extensions import override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.dsi.manifest_object_lookup import ManifestObjectLookup
from metricflow_semantics.experimental.semantic_graph.builder.categorical_dimension_subgraph import (
    CategoricalDimensionSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.builder.entity_join_subgraph import EntityJoinSubgraphGenerator
from metricflow_semantics.experimental.semantic_graph.builder.entity_key_subgraph import (
    EntityKeySubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.builder.metric_subgraph import ComplexMetricSubgraphGenerator
from metricflow_semantics.experimental.semantic_graph.builder.simple_metric_subgraph import (
    SimpleMetricSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.builder.subgraph_generator import SemanticSubgraphGenerator
from metricflow_semantics.experimental.semantic_graph.builder.time_dimension_subgraph import (
    TimeDimensionSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.builder.time_entity_subgraph import TimeEntitySubgraphGenerator
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import MutableSemanticGraph, SemanticGraph
from metricflow_semantics.helpers.performance_helpers import ExecutionTimer
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


class SemanticGraphBuilder:
    """Convenience class that builds a semantic graph using all subgraph generators.

    This will replace `PartialSemanticGraphBuilder` once the PRs that depend on it are merged.
    """

    _ALL_SUBGRAPH_GENERATORS: AnyLengthTuple[Type[SemanticSubgraphGenerator]] = (
        CategoricalDimensionSubgraphGenerator,
        EntityKeySubgraphGenerator,
        EntityJoinSubgraphGenerator,
        SimpleMetricSubgraphGenerator,
        TimeDimensionSubgraphGenerator,
        TimeEntitySubgraphGenerator,
        ComplexMetricSubgraphGenerator,
    )

    @override
    def __init__(self, manifest_object_lookup: ManifestObjectLookup) -> None:
        self._manifest_object_lookup = manifest_object_lookup
        self._verbose_debug_logs = True

    def _build(self, subgraph_generators: Sequence[Type[SemanticSubgraphGenerator]]) -> SemanticGraph:
        current_graph = MutableSemanticGraph.create()
        for generator in subgraph_generators:
            generation_timer = ExecutionTimer()
            with generation_timer:
                generator_instance = generator(self._manifest_object_lookup)
                start_node_count = len(current_graph.nodes)
                start_edge_count = len(current_graph.edges)
                generated_edges = generator_instance.generate_edges()
                generated_edge_count = len(generated_edges)
                current_graph.add_edges(generated_edges)
                added_node_count = len(current_graph.nodes) - start_node_count
                added_edge_count = len(current_graph.edges) - start_edge_count

            if self._verbose_debug_logs:
                logger.debug(
                    LazyFormat(
                        "Generated subgraph",
                        generator=generator.__name__,
                        duration=generation_timer.total_duration,
                        added_node_count=added_node_count,
                        added_edge_count=added_edge_count,
                        generated_edge_count=generated_edge_count,
                    )
                )

        return current_graph

    def build(  # noqa: D102
        self, subgraph_generators: Sequence[Type[SemanticSubgraphGenerator]] = _ALL_SUBGRAPH_GENERATORS
    ) -> SemanticGraph:
        with ExecutionTimer() as build_timer:
            result_graph = self._build(subgraph_generators)
        logger.info(
            LazyFormat(
                "Generated semantic graph.",
                node_count=len(result_graph.nodes),
                edge_count=len(result_graph.edges),
                duration=build_timer.total_duration,
            )
        )
        return result_graph



from __future__ import annotations

import logging
from functools import cached_property
from typing import Any, Callable, Mapping, Union

from typing_extensions import override

from metricflow_semantics.mf_logging.pretty_print import PrettyFormatDictOption, mf_pformat_dict

logger = logging.getLogger(__name__)


LoggedObject = Any  # type: ignore[misc]


class LazyFormat:
    """Lazily formats the given objects into a string representation for logging.

    The formatting is done when this object is converted to a string to be compatible with the standard `logging`
    library. This is done lazily as the formatting operation can be expensive and allows debug log statements to not
    incur a performance overhead in production.

    Example:
        logger.debug(LazyFormat(lambda: LazyFormat("Found path", start_point=point_0, end_point=point_1)))

        ->

        DEBUG - Found path
            start_point: Point(x=0.0, y=0.0)
            end_point: Point(x=1.0, y=1.0)

    To aid migration of existing log statements, the message can be a function so that f-strings can be lazily
    evaluated as well. This style will be deprecated as log statements are updated.

    Example:
        logger.debug(LazyFormat(lambda: f"Result is: {expensive_function()}"))

        ->

        logger.debug(LazyFormat(lambda: f"Result is: {expensive_function()}")
    """

    def __init__(
        self,
        message_title: Union[str, Callable[[], str]],
        **kwargs: Union[LoggedObject, Callable[[], LoggedObject]],
    ) -> None:
        """Initializer.

        Args:
            message_title: The message title or a function that returns a message title. A message is composed of the
            message title and the message body. For this class, the message body is derived from the values of the
            keyword arguments.
            **kwargs: A dictionary of objects to format to text when `__str__` is called on this object. To allow
            for lazy evaluation of argument values, a callable that takes no arguments can be supplied for a key's
            value.
        """
        self._message_title = message_title
        self._kwargs = kwargs

    @cached_property
    def evaluated_value(self) -> str:
        """Cache the result as `__str__` can be called multiple times for multiple log handlers."""
        return mf_pformat_dict(
            self.evaluated_message_title,
            self.evaluated_kwargs,
            format_option=PrettyFormatDictOption(preserve_raw_strings=True),
        )

    @override
    def __str__(self) -> str:
        return self.evaluated_value

    @cached_property
    def evaluated_kwargs(self) -> Mapping[str, LoggedObject]:
        """Return the `kwargs` that was used to construct this object, evaluating value if `Callable`.

        This is cached for similar reasons as `evaluated_value`.
        """
        evaluated_args = {}
        for arg_name, arg_value in self._kwargs.items():
            if callable(arg_value):
                try:
                    evaluated_args[arg_name] = arg_value()
                    continue
                except Exception:
                    logger.warning(
                        "Got an exception while evaluating an argument (see `extra` for this log record). Since"
                        " this is an error with formatting log output, this should not result in system issues."
                        " However, the exception indicates a bug with how the logging call is made and should be"
                        " investigated.",
                        exc_info=True,
                        extra={"arg_name": arg_name, "arg_value": arg_value},
                    )

            evaluated_args[arg_name] = arg_value
        return evaluated_args

    @cached_property
    def evaluated_message_title(self) -> str:
        """Return the `message_title` that was used to construct this object, evaluating value if `Callable`.

        This is cached for similar reasons as `evaluated_value`.
        """
        if callable(self._message_title):
            # May be helpful to wrap this with a try/except as similar to `evaluated_kwargs`.
            return self._message_title()
        else:
            return self._message_title
